/*!
  Reads the configuration file and stores the config parameters to provide quick access
*/

#ifndef CONFIGURATION_H
#define CONFIGURATION_H

#include <fstream>
#include <string>
#include <iostream>

using namespace std;

namespace conf {

    //! Configuration file handler class
    class Conf {
    protected:
        //! Configuration file name
        string filename;

    public:
        //! Default Constructor
        Conf()
        {
            filename = "";
        }

        //! Constructor with filename as the parameter
        Conf(string f)
        {
            filename = f;
            loadConf();
        }

        //! Set the filename
        void setFilename(string f)
        {
            filename = f;
        }

        //! Get the filename
        string& getFilename()
        {
            return filename;
        }

        //! Read the default configuration file and extracts configuration parameters
        int loadConf()
        {   return loadConf(filename); }

        //! Read the configuration file and extracts configuration parameters
        virtual int loadConf(string f) = 0;

        //! Save the configuration into a file
        virtual int saveConf(string f) = 0;

    };





    /*!
        System configuration file class for VisApp
        Internal parameters required for VisApp are loaded and stored
    */
    class VisAppSysConf : public Conf {
    public:
        //! Default constructor
        VisAppSysConf() : Conf() {}

        //! Constructor
        VisAppSysConf(string f) : Conf(f) {}

        //! Read the configuration file and extracts configuration parameters
        /*!
            Return value: 0 if successful, -1 otherwise
        */
        int loadConf(string f) {
            filename = f;
            if (filename=="") {
                cerr<<"Filename is not defined"<<endl;
                return -1;
            }
            string line;
            ifstream myfile;
            //cout<<"Filename: "<<filename.c_str()<<endl;
            myfile.open(filename.c_str(),ios::in);
            if (myfile.is_open()) {
                string myconf, value;
                myfile >> myconf;
                while(!myfile.eof()) {
                    if (!myconf.compare("#")) {
                        getline(myfile,line);
                    }
                    else {
                        getline(myfile,line);
                        cout<<myconf<<" : "<<line<<endl;
                    }
                    myfile >> myconf;
                }
                myfile.close();
                return 0;
            }
            else {
                cerr<<"Unable to open file: "<<filename<<endl;
                return -1;
            }
        }

        //! Save the configuration into a file
        /*!
            Return value: 0 if successful, -1 otherwise
        */
        int saveConf(string f) {
            if (f=="") {
                cerr<<"Filename is not defined"<<endl;
                return -1;
            }
            string line;
            ofstream myfile;
            //cout<<"Filename: "<<f.c_str()<<endl;
            myfile.open(f.c_str(),ios::out);
            if (myfile.is_open()) {

                myfile.close();
                return 0;
            }
            else {
                cerr<<"Unable to open file: "<<f<<endl;
                return -1;
            }
        }
    };






    /*!
        Configuration file class for VisApp
        Parameters for mapping, localization, camera etc are handled by this class
    */
    class VisAppConf : public Conf {
    public:
        int TRACKINGMODE;   //tracking mode show if there is temporal coherence between consecutive images
        int FEATUREDIM;     // dimension of the feature vector

        int FMATSAMPLES;
        int FMATMININLIERS;

        int MAXMATCHES;    // maximum number of feature matches to accept the image as a candidate match (above this limit they are too similar)
        int MINMATCHES;    // minimum number of valid feature matches to accept the image as a candidate match and try to insert the image to the map. Less than that dont even try to collect the votes, just ignore
        int MAXVERTEXDEGREE;   // maximum number of edges a vertex can have

        string CLUSTER_FILE;
        string INDEX_FILE;

        string GRIDMAP_FILE;

        //! Default constructor
        VisAppConf() : Conf(){
            setDefaultParameters();
        }

        VisAppConf(string fname) : Conf(){
            setDefaultParameters();
            loadConf(fname);
        }


        void setDefaultParameters() {
            TRACKINGMODE = 1;
            FEATUREDIM = 128;

            FMATSAMPLES = 8;
            FMATMININLIERS = 10;

            MAXMATCHES = 25;
            MINMATCHES = 15;
            MAXVERTEXDEGREE = 10;

            CLUSTER_FILE = "/home/gorkem/code/data/all_noGames.cluster_50K_100i_F.hdf5";
            INDEX_FILE = "/home/gorkem/code/data/all_noGames.cluster_50K_100i_F.hdf5.index.hdf5";

            GRIDMAP_FILE = "sys/gridmap.png";
        }


        //! Read the configuration file and extracts configuration parameters
        /*!
            Return value: 0 if successful, -1 otherwise
        */
        int loadConf(string f) {

            filename = f;
            if (filename=="") {
                cerr<<"Filename is not defined"<<endl;
                return -1;
            }
            //cout<<"Reading configuration file: "<<f<<endl;

            ifstream myfile;
            //cout<<"Filename: "<<filename.c_str()<<endl;
            myfile.open(filename.c_str(),ios::in);
            if (myfile.is_open()) {
                istringstream parse;
                string line, temp, valuestr;
                int value;

                while(getline(myfile,line)) {
                    parse.clear();
                    parse.str(line);
                    parse >> temp;

                    if(temp=="FMATSAMPLES") {
                        parse >> temp >> value;
                        FMATSAMPLES = value;
                    }
                    else if(temp=="FMATMININLIERS") {
                        parse >> temp >> value;
                        FMATMININLIERS = value;
                    }
                    else if(temp=="MAXMATCHES") {
                        parse >> temp >> value;
                        MAXMATCHES = value;
                    }
                    else if(temp=="MINMATCHES") {
                        parse >> temp >> value;
                        MINMATCHES = value;
                    }
                    else if(temp=="FEATUREDIM") {
                        parse >> temp >> value;
                        FEATUREDIM = value;
                    }
                    else if(temp=="TRACKINGMODE") {
                        parse >> temp >> value;
                        TRACKINGMODE = value;
                    }
                    else if(temp=="CLUSTER_FILE") {
                        parse >> temp >> valuestr;
                        CLUSTER_FILE = valuestr;
                    }
                    else if(temp=="INDEX_FILE") {
                        parse >> temp >> valuestr;
                        INDEX_FILE = valuestr;
                    }
                    else if(temp=="GRIDMAP_FILE") {
                        parse >> temp >> valuestr;
                        GRIDMAP_FILE = valuestr;
                    }
                    else {
                        continue;
                    }
                }
                myfile.close();
                return 0;
            }
            else {
                cerr<<"Unable to open file: "<<filename<<endl;
                return -1;
            }
        }

        //! Save the configuration into a file
        /*!
            Return value: 0 if successful, -1 otherwise
        */
        int saveConf(string f) {
            if (f=="") {
                cerr<<"Filename is not defined"<<endl;
                return -1;
            }
            ofstream myfile;
            //cout<<"Filename: "<<f.c_str()<<endl;
            myfile.open(f.c_str(),ios::out);
            if (myfile.is_open()) {

                myfile << "The syntax is as follows: KEYWORD = VALUE";
                myfile << "One keyword-value pair per line";
                myfile << "Anything else will be considered as comment and will be ignored";
                myfile << "FMATSAMPLES = " << FMATSAMPLES << endl;
                myfile << "FMATMININLIERS = " << FMATMININLIERS <<endl;
                myfile << "MAXMATCHES = " << MAXMATCHES << endl;
                myfile << "MINMATCHES = " << MINMATCHES << endl;
                myfile << "FEATUREDIM = " << FEATUREDIM << endl;
                myfile << "TRACKINGMODE = " << TRACKINGMODE << endl;
                myfile << "CLUSTER_FILE = " << CLUSTER_FILE << endl;
                myfile << "INDEX_FILE = " << INDEX_FILE << endl;
                myfile << "GRIDMAP_FILE = " << GRIDMAP_FILE <<endl;

                myfile.close();
                return 0;
            }
            else {
                cerr<<"Unable to open file: "<<f<<endl;
                return -1;
            }
        }

    };
}
#endif // CONFIGURATION_H
