/***************************************************************
 *
 * Written by: Gorkem Erinc, gerinc@gmail.com
 *
 * Description:
 *    Dictionary structure in Bag of Words integrated in
 *    appearance-based maps.
 *
 ***************************************************************/

#include "dictionary.h"
#include <sstream>
#include <opencv2/core/core.hpp>

using namespace std;
using namespace converter;
using namespace boost;


Dictionary::Dictionary(string clusterFile, string indexFile)
{
    index = new cvflannIndex::Index<FT,CT>();
    index->loadClusters(clusterFile);
    index->loadClusterIndex(indexFile);
}

Dictionary::~Dictionary()
{
    clear();
    delete index;
    index = NULL;
}


void Dictionary::clear() {
    BOOST_FOREACH(InvIndex::value_type &w, invIndex) {
        BOOST_FOREACH(OccuranceList::value_type &o, w.second) {
            o.second.clear();
        }
        w.second.clear();
    }
    invIndex.clear();
    //dict.clear();
}




void Dictionary::queryDictionary(Image *img, Mat &indices, Mat &distances) {

    int N = (int)img->features.size();

    Mat query(N,appConf.FEATUREDIM,cv::flann::CvType<CT>().type());
    for(int i=0; i<N; i++) {
        for(int j=0; j<appConf.FEATUREDIM; j++) {
            query.at<CT>(i,j) = (CT)img->features[i]->Descr().Data()[j];
        }
    }

    int nn = 1;
    indices.create(query.rows,nn,CV_32S);
    distances.create(query.rows,nn,CV_32F);

    index->getIndex()->knnSearch(query, indices, distances, nn, cv::flann::SearchParams(128));

}


void Dictionary::addImage(Image *img) {
    int imgIndex = img->getIndex();

    bool inserted = false;
    if(img->words.size()>0)
        inserted = true;

    // This if check is added for performance purposes...
    // It comes into play when an image already inserted into some other map is being inserted into this one
    // (word list is filled with the word indices pointing to words in its own dictionary)
    // Assuming we keep the dictionaries fixed, there is no need to query the dictionary again
    // However, this check can be removed to generalize the process
    if(!inserted) {

        if(img->features.size()==0)
            img->extractFeatures();

        int N = (int)img->features.size();


        // Compute nearest neighbors
        Mat indices;
        Mat distances;
        queryDictionary(img, indices, distances);

        // Add to the matched word's occurances to inverted index
        for(int i=0; i<N; i++) {

            double x = ((feat::FeatProp *)(img->features[i]->prop))->Location().X();
            double y = ((feat::FeatProp *)(img->features[i]->prop))->Location().Y();
            cv::Point2d pnew(x,y);

            int wordIndex = indices.at<int>(i,0);
            invIndex[wordIndex][imgIndex].push_back(pnew);
            img->words.push_back(make_pair(wordIndex,pnew));
        }
    }
    else {      
        // Add to the matched word's occurances to inverted index
        for(int i=0; i<img->words.size(); i++) {
            invIndex[img->words[i].first][imgIndex].push_back(img->words[i].second);
        }
    }
}


// Searches the dictionary using the image and returns the matching feature-matches in matches
void Dictionary::findMatches(Image *img, PointMatchMap &matches)
{
    int N = (int)img->features.size();
    int M = (int)img->words.size();

    if(M>0) {
        // directly use the word indices to find the matches
        for(int i=0; i<M; i++) {
            BOOST_FOREACH(OccuranceList::value_type &o, invIndex[img->words[i].first]) {
                BOOST_FOREACH(cv::Point2d &point, o.second) {
                    matches[o.first].first.push_back(img->words[i].second);
                    matches[o.first].second.push_back(point);
                }
            }
        }
    }
    else {
        // -- Compute nearest neighbors
        Mat indices;
        Mat distances;
        queryDictionary(img, indices, distances);

        // for each feature
        for(int i=0; i<N; i++) {

            double x1 = ((FeatProp *)(img->features[i]->prop))->Location().X();
            double y1 = ((FeatProp *)(img->features[i]->prop))->Location().Y();
            cv::Point2d p1(x1,y1);

            if(invIndex.find(indices.at<int>(i,0)) != invIndex.end()) {
                BOOST_FOREACH(OccuranceList::value_type &o, invIndex[indices.at<int>(i,0)]) {
                    BOOST_FOREACH(cv::Point2d &point, o.second) {
                        cv::Point2d p2(point.x,point.y);
                        matches[o.first].first.push_back(p1);
                        matches[o.first].second.push_back(p2);
                    }
                }
            }
        }
    }
}





void Dictionary::print() {
    Debug(0) << "Printing the occurances in the dictionary of size: "<<invIndex.size() << endl;
    BOOST_FOREACH(InvIndex::value_type &ol, invIndex) {
        Debug(0)<<"Word: "<<ol.first<<endl;
        BOOST_FOREACH(OccuranceList::value_type &o, ol.second) {
            BOOST_FOREACH(cv::Point2d &point, o.second) {
                Debug(0)<<"Appeared in: "<<o.first<<" at "<<point.x<<" "<<point.y<<endl;
            }
        }
    }
}


// This function actually writes inverted index file
void Dictionary::writeDictionary(string filename) {
    writeInvIndex(filename);
}

void Dictionary::writeInvIndex(string filename) {

    Debug(1)<<"Writing database to file: "<<filename<<endl;

    ofstream myfile;
    myfile.open(filename.c_str(),ios::out);
    if(myfile.is_open()) {
        myfile << "Occupied dictionary size: " << invIndex.size() << endl;

        BOOST_FOREACH(InvIndex::value_type &ol, invIndex) {
            myfile << "Word: " << ol.first << endl;
            BOOST_FOREACH(OccuranceList::value_type &o, ol.second) {
                BOOST_FOREACH(cv::Point2d &point, o.second) {
                    myfile << "Appeared in image: " << o.first<<" at "<<point.x<<" "<<point.y<<endl;
                }
            }
        }
        myfile.close();
    }
}



void Dictionary::readDictionary(string filename, ImageTable &imageLookup) {
    readInvIndex(filename, imageLookup);
}

void Dictionary::readInvIndex(string filename, ImageTable &imageLookup) {

    Debug(1)<<"Reading database from file: "<<filename<<endl;

    ifstream myfile(filename.c_str());
    try {
        if(!myfile.is_open())
            throw 1;

        clear();

        string line,temp;
        istringstream parse;
        getline(myfile,line); //  number of words in the dict file

        Debug(1) << "\tStarting reading " << line << " words" << endl;


        int wordIndex = -1;
        int imageIndex = -1;

        while(getline(myfile,line)) {
            if(!converter::isEmpty(line)) {

                parse.clear();
                parse.str(line);
                parse >> temp;

                if(temp=="Word:") {
                    parse >> wordIndex;
                }
                else if(temp=="Appeared") {
                    if(wordIndex==-1) {
                        throw 2;
                    }
                    parse >> temp >> temp >> imageIndex;
                    parse >> temp;
                    double x,y;
                    parse >> x >> y;
                    cv::Point2d p(x,y);
                    invIndex[wordIndex][imageIndex].push_back(p);
                    imageLookup[imageIndex]->words.push_back(make_pair(wordIndex,p));
                }
            }
        }

        myfile.close();
    }
    catch(int errId) {
        switch (errId) {
        case 1:
            cout<<"ERROR: Database file could not be opened" <<endl;
            break;
        case 2:
            cout<<"ERROR: Occurance with no index detected" <<endl;
            break;
        }
        exit(0);
    }
}


