#ifndef IMAGE_H
#define IMAGE_H

#include <list>
#include <vector>
#include <sys/time.h>
#include "defs.h"
#include "feature.h"
#include "graph.h"
#include "imageutils.h"
#include <boost/unordered_map.hpp>

using namespace feat;

typedef cv::Point3d Pose;


class Image
{
public:
    enum ImageFormat {
        UNKNOWN = -1,
        GRAY = 1,
        RGB = 3
    };

    enum FileFormat {
        PPM = 1,
        PGM = 2
    };
private:
    //! Image data
    unsigned char* data;

    //! Image signiture ID
    int index;

    string title;

    string fileName;
    FileFormat fileFormat;

    //! Timestamp of the image
    struct timeval timestamp;

    //! Metric location where the image is taken (to be used for visualization)
    Pose pose;

    int width, height;
    ImageFormat imageFormat;

public:
    //! Default constructor
    Image();

    //! Constructor
    Image(unsigned char*, int w, int h, ImageFormat f, Pose p = Pose());

    //! Copy constructor
    Image(Image* img);

    //! Destructor
    ~Image();

    unsigned char* getImageData();
    void setImageData(unsigned char*, int w, int h, ImageFormat f, Pose p = Pose());
    void deleteData();

    void setIndex(int i) { index = i; }
    int getIndex() { return index; }

    void setTitle(string t) { title=t; }
    string getTitle() { return title; }

    struct timeval getTimestamp() { return timestamp; }
    void setTimestamp(struct timeval t) { timestamp = t; }

    Pose getPose() { return pose; }
    void setPose(Pose p ) { pose = p; }

    bool isEmpty() { if(data==NULL) return true; return false;}

    string getFilename() { return fileName; }
    void setFilename(string fname) { fileName = fname; }
    FileFormat getFileFormat() { return fileFormat; }
    void setFileFormat(int ff) { fileFormat = (FileFormat)ff; }
    void setFileFormat(FileFormat ff) { fileFormat = ff; }

    int getWidth() { return width; }
    int getHeight() { return height; }
    int getSize() { return width*height*imageFormat; }

    ImageFormat getImageFormat() { return imageFormat; }
    string getImageFormatString();

    int loadImageFromFile();
    int loadImage(string filename, FileFormat fileformat);
    int saveImage(string filename, FileFormat fileformat);

    void resize();

    int extractFeatures();
    int numFeatures() {
        if(features.size()>0) return features.size();
        return words.size();
    }
    void matchImage(Image *img, PointMatchList &pl);

    void printFeatures();
    void printFeatures(string filename);
    void printWords();

    //! List of features extracted from the image stored locally
    vector<Feat*> features;

    //! List of words matched in the dictionary
    vector<pair<wordIndex,cv::Point2d> > words;

    //! Pointer to the vertex that contains this image
    vertexDescriptor vertex;
};

//! For each access to images stored in the appearance-based map
typedef boost::unordered_map<imageIndex,Image*> ImageTable; // map<imageindex, image*>


#endif // IMAGE_H
