/***************************************************************
 *
 * Written by: Gorkem Erinc, gerinc@gmail.com
 *
 * Description:
 *    Image utility class provides functions to
 *    create, modify, and visualize appearance-based maps.
 *
 ***************************************************************/

#include "imageutils.h"
#include <iostream>
#include <stdio.h>
#include "index.h"
#include <set>

using namespace std;
using namespace cv;

namespace iu {

    // The number of features in the list doesn't reflect the correct unique features in each image
    // since some features are used more than once
    int getUniqueFeats(PointMatchList &pointList)
    {
        PointList *p1 = &pointList.first;
        PointList *p2 = &pointList.second;

        map<pair<int,int>,int> checklist1,checklist2;
        map<pair<int,int>,int>::iterator it;

        for(unsigned int i=0; i<p1->size(); i++) {
            checklist1[make_pair(floor((*p1)[i].x),floor((*p1)[i].y))]++;
        }
        for(unsigned int i=0; i<p2->size(); i++) {
            checklist2[make_pair(floor((*p2)[i].x),floor((*p2)[i].y))]++;
        }

        int c1=0,c2=0;
        for(it=checklist1.begin();it!=checklist1.end();it++)
            c1++;
        for(it=checklist2.begin();it!=checklist2.end();it++)
            c2++;

        return min(c1,c2);
    }


    int	enforceFundMatConstraint(Mat &F, PointMatchList &matchList, int filterlevel, int img2Height, IplImage *img1, IplImage *img2)
    {
        int method = 2;
        int errMethod = 0;
        bool ransac = false;// ransac function I coded
        int numiter = 10000; // for ransac

        Debug(0)<<"\t#allmatches: "<<matchList.first.size()<<endl;

        removeDuplicates(matchList);
        Debug(0)<<"\t#duplicate removal: "<<matchList.first.size()<<endl;
        visualizeMatches(img1,img2,matchList,"removeDup");
        if((int)matchList.first.size() < appConf.MINMATCHES) {
            Debug(0) << "Stopping early due to low number of matches" << endl;
            return matchList.first.size();
        }

        int unique = getUniqueFeats(matchList);
        if(unique < appConf.MINMATCHES) {
            Debug(0) << "Stopping early due to low number of unique matches: "<<unique<< endl;
            //return unique;
        }

        if(filterlevel & 0b001) {
            struct timeval tStart,tEnd;
            gettimeofday(&tStart,NULL);

            if(ransac)
                F = calcFundamentalMat_Ransac(&matchList, numiter, method, appConf.FMATSAMPLES, appConf.FMATMININLIERS);
            else
                F = calcFundamentalMat(&matchList, method);

            gettimeofday(&tEnd,NULL);
            double elapsed = (tEnd.tv_sec -tStart.tv_sec) + (tEnd.tv_usec - tStart.tv_usec)/1000000.0;
            Debug(0)<<"\t\t\t\t**********FindfundamentalMat: "<<elapsed<<endl;



            if(F.rows==0)
                return 0;

            calcFundMatError(F, matchList, errMethod);
            Debug(0)<<"\t#inliers (F support): "<<matchList.first.size()<<endl;
            visualizeMatches(img1,img2,matchList,"inliers - Residual");

            if((int)matchList.first.size() < appConf.MINMATCHES) {
                Debug(0) << "Stopping early due to low number of matches" << endl;
                return matchList.first.size();
            }
        }

        if(filterlevel & 0b010) {
            removeOutliers(matchList);
            Debug(0)<<"\t#inliers (after outlier removal): "<<matchList.first.size()<<endl;
            visualizeMatches(img1,img2,matchList,"inliers - after outlier removal");

            if((int)matchList.first.size() < appConf.MINMATCHES) {
                Debug(0) << "Stopping early due to low number of matches" << endl;
                return matchList.first.size();
            }
        }

        if(filterlevel & 0b100) {
            //int imgHeight = img2->height;
            img2Height = 240;
            if(img2Height==0) {
                Debug(0)<<"WARNING: Current implementation of spatial consistency check requires the height of image2"<<endl;
                Debug(0)<<"WARNING: Skipping spatial consistency check!"<<endl;
            }
            else {
                applySpatialConsistency(matchList, img2Height);
                Debug(0)<<"\t#inliers (final - after spatial consistency check): "<<matchList.first.size()<<endl;
                visualizeMatches(img1,img2,matchList,"inliers - spatialCons");
            }
        }


        // eliminate duplicates
        // compute F
        // - ransac vs no ransac
        // - F algorithms
        // - error functions
        // Compute err & inliers
        // remove outliers
        // spatial consts


        unique = getUniqueFeats(matchList);
        //cout<<"function unique: "<<unique<<" matchsize: "<<matchList.first.size()<<" with min pass "<<appConf.MINMATCHES<<endl;
        if(unique < appConf.MINMATCHES) {
            Debug(0) << "Stopping early due to low number of unique matches: "<<unique<< endl;
            //return unique;
        }


        return matchList.first.size();
    }


    // Methods:
    // 0: 7 point (opencv)
    // 1: 8 point (opencv)
    // 2: ransac (opencv)
    // 3: LMedS (opencv)
    // 4: Torr (EGT - Matlab)
    // INPUT: p1(Nx2), p2(Nx2)
    Mat calcFundamentalMat(PointMatchList *pointList, int method)
    {        
        Mat F;
        Mat p1(pointList->first);
        Mat p2(pointList->second);

        if (method==0) {
            F = findFundamentalMat(p1,p2,CV_FM_7POINT);
        }
        else if  (method==1) {
            F = findFundamentalMat(p1,p2,CV_FM_8POINT);
        }
        else if (method==2) {
            F = findFundamentalMat(p1,p2,CV_FM_RANSAC,1.0,0.9999);
        }
        else if (method==3) {
            F = findFundamentalMat(p1,p2,CV_FM_LMEDS,1.0,0.99);
        }
        else {// (method==4) {
            //F = findFundamentalMat_byTorr(p1,p2);
        }
        forceRank2(F);

        return F;
    }


    Mat calcFundamentalMat_Ransac(PointMatchList *pointList, int numiter, int method, int numsamples, int mininliers) {
        /*
        FOR_LOOP
            points = select random points
            model = compute model using these random points
            supporters = compute supporters of this model
            if supporters > bestsupporters
                modelnew = compute model using these supporters
                error = compute error for this model with these supporters
                if error < besterror
                    store them all
                endif
            endif
        END
        */

        int numPoints = pointList->first.size();
        double errbest = std::numeric_limits<double>::infinity();
        int maxInliers = 0;
        Mat Fbest;

        if (numPoints < numsamples) {
            return Fbest;
        }

        // initialize random seed:
        srand ( time(NULL) );

        for(int i=0; i<numiter; i++)
        {
            //cout<<i<<endl;

            // choose random samples
            set<int> randindices;
            while(randindices.size()!=(unsigned int)numsamples) {
                int sel = floor(rand() % numPoints);
                if(randindices.find(sel)==randindices.end()) { //not selected before
                    randindices.insert(sel);
                }
            }
            PointMatchList randsamples;
            for(set<int>::iterator it=randindices.begin(); it!=randindices.end(); it++) {
                randsamples.first.push_back(pointList->first[*it]);
                randsamples.second.push_back(pointList->second[*it]);
            }

            // Compute F matrix for this subset
            Mat F = calcFundamentalMat(&randsamples, method);

            // Take a backup of pointlist
            PointMatchList inlierList;
            for(unsigned int j=0; j<pointList->first.size(); j++) {
                inlierList.first.push_back(pointList->first[i]);
                inlierList.second.push_back(pointList->second[i]);
            }

            // Compute inliers and associated error for ALL points
            calcFundMatError(F, inlierList);

            if(inlierList.first.size() > (unsigned int)mininliers)
            {   // implies that we may have found a good model that fits the underlying data

                // Compute F matrix for all inliers
                F = calcFundamentalMat(&inlierList, method);
                double err = calcFundMatError(F, inlierList);
                if((int)inlierList.first.size() > maxInliers || ((int)inlierList.first.size() == maxInliers && err < errbest)) {
                    Fbest = F;
                    errbest = err;
                    maxInliers = inlierList.first.size();
                    //cout<<"best found: "<<maxInliers<<" with "<<errbest<<endl;
                }
            }
        }

        return Fbest;
    }



    double calcFundMatError(Mat &F, PointMatchList &pointList, int errMethod, double errThreshold) {
        if(errMethod == 0) {
            if(errThreshold > 0)
                return calcFundMatError_Resudial(F,pointList,errThreshold);
            else
                return calcFundMatError_Resudial(F,pointList);
        }
        else {
            if(errThreshold > 0)
                return calcFundMatError_Sampsons(F,pointList,errThreshold);
            else
                return calcFundMatError_Sampsons(F,pointList);
        }
    }


    double calcFundMatError_Resudial(Mat &F, PointMatchList &pointList, double errThreshhold) {

        /* FROM MATLAB
        function [err dist1 dist2] = disterrf( F, x, xp )
        x(:,3) = 1;
        xp(:,3) = 1;
        L1 = normalizeLine(F * x');
        L2 = normalizeLine(F' * xp');
        dist1 = abs(dot( xp', L1 ));
        dist2 = abs(dot( L2, x' ));
        err = sum( dist1.^2 + dist2.^2 ) / size(x,1);
        */

        Mat points1(pointList.first);
        Mat points2(pointList.second);
        Mat p1 = Mat(points1.reshape(1).t()).clone();
        Mat p2 = Mat(points2.reshape(1).t()).clone();

        pointList.first.clear();
        pointList.second.clear();

        // Convert to homogenous
        if(p1.rows==2) {
            p1.push_back(Mat(1,p1.cols,CV_64F,Scalar(1.0)));
        }
        if(p2.rows==2) {
            p2.push_back(Mat(1,p2.cols,CV_64F,Scalar(1.0)));
        }

        Mat L1 = F*p1;
        Mat L2 = F.t()*p2;
        L1 = normalizeLine(L1);
        L2 = normalizeLine(L2);

        // dist2 = abs(dot( L2, x' ));
        // diag(xp*L1)'
        Mat dist1 = cv::abs((p2.t()*L1).diag());
        Mat dist2 = cv::abs((L2.t()*p1).diag());

        double error = 0;
        // Returns the error for all points OR returns the error only for inliers
        // For all points
        int numPoints = dist1.rows;
        for(int k=0; k<numPoints; k++) {
            double d1 = dist1.at<double>(k,0);
            double d2 = dist2.at<double>(k,0);
            error = pow(d1,2) + pow(d2,2);
            if(d1 < errThreshhold && d2 < errThreshhold) {
                cv::Point2d pt1(p1.at<double>(0,k), p1.at<double>(1,k));
                cv::Point2d pt2(p2.at<double>(0,k), p2.at<double>(1,k));
                pointList.first.push_back(pt1);
                pointList.second.push_back(pt2);
            }
        }
        return error;
    }


    double calcFundMatError_Sampsons(Mat &F, PointMatchList &pointList, double errThreshhold) {
        /* FROM MATLAB
        %  x: n x 3 vector
        %  xp: n x 3 vector
        function err = sampsonerrf( F, x, xp )
        x(:,3) = 1;
        xp(:,3) = 1;
        L = F * x';
        Lp = F' * xp';
        num = sum(xp' .* L).^2;
        den = L(1,:).^2 + L(2,:).^2 + Lp(1,:).^2 + Lp(2,:).^2;
        err = sum(num ./ den) / size(x,1);
        */

        Mat points1(pointList.first);
        Mat points2(pointList.second);
        Mat p1 = Mat(points1.reshape(1).t()).clone();
        Mat p2 = Mat(points2.reshape(1).t()).clone();

        pointList.first.clear();
        pointList.second.clear();


        // Convert to homogenous
        if(p1.rows==2) {
            p1.push_back(Mat(1,p1.cols,CV_64F,Scalar(1.0)));
        }
        if(p2.rows==2) {
            p2.push_back(Mat(1,p2.cols,CV_64F,Scalar(1.0)));
        }

        Mat L = F*p1;
        Mat Lp = F.t()*p2;

        Mat t1 = p2.mul(L); // p2.*L
        Mat t2(3,1,CV_64F,Scalar(1));
        Mat t3 = t1.t()*t2; // t3(Nx1) = sum(t1=p2.*L)
        Mat num = t3.mul(t3); // num = sum(p2.*L)^2
        Mat den = L.row(0).mul(L.row(0)) + L.row(1).mul(L.row(1)) + Lp.row(0).mul(Lp.row(0)) + Lp.row(1).mul(Lp.row(1));
        Mat errors = num.mul(1/den.t()); //errors(Nx1)
        Scalar err = sum(errors);

        int numPoints = errors.rows;
        for(int k=0; k<numPoints; k++) {
            if(errors.at<double>(k,0) < errThreshhold) {
                cv::Point2d pt1(p1.at<double>(0,k), p1.at<double>(1,k));
                cv::Point2d pt2(p2.at<double>(0,k), p2.at<double>(1,k));
                pointList.first.push_back(pt1);
                pointList.second.push_back(pt2);
            }
        }

        return err(0);
    }




    Mat normalizeLine(Mat &L) {
        //n = sqrt( L(1,:).^2 + L(2,:).^2 );
        Mat n;
        magnitude(L.row(0),L.row(1),n);

        //Lhat = L ./ repmat( n, 3, 1 );
        return L / repeat(n,3,1);
    }


    void forceRank2(Mat &F) {
        if(F.cols!=3 || F.rows!=3) {
            printf("Dimension is not correct\n");
        }
        else {
            SVD svd(F,SVD::FULL_UV);
            svd.w.at<double>(2,0) = 0.0;
            F = svd.u*Mat::diag(svd.w)*svd.vt;
        }
    }



    void removeDuplicates(PointMatchList &pointList)
    {
        PointList *p1 = &pointList.first;
        PointList *p2 = &pointList.second;

        if(p1->size() != p2->size()) {
            cout<<"ERROR: Unmatched dimensions in removeDublicates inputs"<<endl;
            exit(0);
        }

        for(unsigned int i=0; i<p1->size(); i++) {
            for(unsigned int j=i+1; j<p1->size(); j++) {
                if(areSame((*p1)[i].x,(*p1)[j].x) &&
                   areSame((*p1)[i].y,(*p1)[j].y) &&
                   areSame((*p2)[i].x,(*p2)[j].x) &&
                   areSame((*p2)[i].y,(*p2)[j].y))
                {
                    Debug(0) << "Duplicate found: ("<<(*p1)[i]<<")-("<<(*p2)[i]<<")"<<endl;
                    p1->erase(p1->begin()+j);
                    p2->erase(p2->begin()+j);
                }
            }
        }
    }


    void removeOutliers(PointMatchList &matchList)
    {
        PointList *p1 = &matchList.first;
        PointList *p2 = &matchList.second;

        if(p1->size() != p2->size()) {
            cout<<"ERROR: Unmatched dimensions in removeOutliers inputs"<<endl;
            exit(0);
        }

        int runs = 0;
        int totalremoved = 0;
        int removed = 0;
        do {
            runs++;
            int N = p1->size();
            double dist[N];
            double distmean = 0;

            //compute dist mean
            for(int i=0; i<N; i++) {
                int x1 = (*p1)[i].x;
                int y1 = (*p1)[i].y;
                int x2 = (*p2)[i].x;
                int y2 = (*p2)[i].y;
                dist[i] = sqrt(pow(x1-x2,2)+pow(y1-y2,2));
                distmean += dist[i];
            }
            distmean = distmean/N;

            // compute dist deviation
            double diststddev = 0;
            for(int i=0; i<N; i++) {
                diststddev += pow(dist[i]-distmean,2);
            }
            diststddev = sqrt(diststddev/(N-1));

            //cout<<"dists:"<<endl;
            removed = 0;
            for(int i=N-1; i>=0; i--) {
                //cout<<dist[i];

                if(isOutlier(dist[i],distmean,diststddev,N)) {
                    p1->erase(p1->begin()+i);
                    p2->erase(p2->begin()+i);
                    removed++;
                    //cout<< " removed ";
                }
                //cout<< "  with "<<prob*N<<endl;
            }
            totalremoved += removed;

        } while(removed > 0);
        Debug(0)<<totalremoved<<" outliers are removed in "<<runs<<" runs"<<endl;
    }


    void applySpatialConsistency(PointMatchList &matchList, int imgHeight, int nn) {

        int N = matchList.first.size();

        Mat p1(0,2,CV_64F);
        Mat p2(0,2,CV_64F);
        map<int,int> indexMap1;
        map<int,int> indexMap2;
        for(int i=0; i<N; i++) {
            bool found = false;
            int j=0;
            for(j=0; j<p1.rows; j++) {
                if(areSame(p1.at<double>(j,0),matchList.first[i].x) && areSame(p1.at<double>(j,1),matchList.first[i].y)) {
                    found = true;
                    break;
                }
            }
            if(!found) {
                Mat t1 = Mat(matchList.first[i]).clone();
                t1 = t1.t();
                p1.push_back(t1);
                indexMap1[i] = p1.rows-1;
            }
            else
                indexMap1[i] = j;
        }
        for(int i=0; i<N; i++) {
            bool found = false;
            int j=0;
            for(j=0; j<p2.rows; j++) {
                if(areSame(p2.at<double>(j,0),matchList.second[i].x) && areSame(p2.at<double>(j,1),matchList.second[i].y)) {
                    found = true;
                    break;
                }
            }
            if(!found) {
                Mat t2 = Mat(matchList.second[i]).clone();
                t2 = t2.t();
                p2.push_back(t2);
                indexMap2[i] = p2.rows-1;
            }
            else
                indexMap2[i] = j;
        }


        nn++;
        if(p1.rows<nn || p2.rows<nn) {
            Debug(0) << "Not enough points to compute spatial neighbors" << endl;
            matchList.first.clear();
            matchList.second.clear();
        }
        else {
            Mat indices1(p1.rows,nn,CV_32S);
            Mat indices2(p2.rows,nn,CV_32S);
            Mat dists1(p1.rows,nn,CV_32F);
            Mat dists2(p2.rows,nn,CV_32F);


            cvflannIndex::Index<double,double> idx1;
            idx1.setClusters(p1);
            idx1.buildClusterIndex();
            idx1.getIndex()->knnSearch(p1, indices1, dists1, nn, cvflann::SearchParams(128));

            cvflannIndex::Index<double,double> idx2;
            idx2.setClusters(p2);
            idx2.buildClusterIndex();
            idx2.getIndex()->knnSearch(p2, indices2, dists2, nn, cvflann::SearchParams(128));


            int minneighbors = ceil((nn-1)*0.5);
            Debug(0) <<"min:"<<minneighbors<<" out of "<<nn-1<<endl;


            int count = 0;
            for(int i=N-1; i>=0; i--) {
                // checking if matchList[i] is a valid match

                vector<int> seenPoints;
                int indx1 = indexMap1[i]; // index in unique point list p1
                for(int j=1; j<nn; j++) { // j=0 is yourself
                    int m = indices1.at<int>(indx1,j);
                    for(int k=0; k<N; k++) {
                        if(indexMap1[k]==m)
                            seenPoints.push_back(indexMap2[k]);
                    }
                }
                int indx2 = indexMap2[i]; // index in unique point list p2
                for(int j=1; j<nn; j++) { // j=0 is yourself
                    int m = indices2.at<int>(indx2,j);
                    bool found = false;
                    for(int k=0; k<seenPoints.size(); k++) {
                        if(seenPoints[k]==m) {
                            found = true;
                            break;
                        }
                    }
                    if(found) {
                        count++;
                    }
                }
                if(count < minneighbors) {
                    matchList.first.erase(matchList.first.begin()+i);
                    matchList.second.erase(matchList.second.begin()+i);
                    Debug(0)<<"Spatial check removed one matching candidate"<<endl;
                }
            }


            /*
            for(int i=p1.rows-1; i>=0; i--) { //p1.rows=p2.rows
                // for each feature  match

                //cout<<"point: ("<<p1.at<double>(i,0)<<" "<<p1.at<double>(i,1)<<") -> ("<<p2.at<double>(i,0)<<" "<<p2.at<double>(i,1)<<")"<<endl;

                cout<<"i:"<<i<<endl;
                cout<<"indices1: "<<indices1.row(i)<<endl;
                cout<<"dist1: "<<dists1.row(i)<<endl;
                cout<<"indices2: "<<indices2.row(i)<<endl;
                cout<<"dist2: "<<dists2.row(i)<<endl;

                Mat angs(0,1,CV_64F);
                for(int j=0; j<closest1.size(); j++) { // index 0 will self-match -> I keep it to compute its ang as well
                    for(int k=0; k<closest2.size(); k++) {
                        if(closest1[j] == closest2[k]) {
                            double x1 = p1.at<double>(closest1[j],0);
                            double y1 = p1.at<double>(closest1[j],1);
                            double x2 = p2.at<double>(closest2[k],0);
                            double y2 = p2.at<double>(closest2[k],1) + imgHeight; // HARDCODED
                            double ang = atan2(y2-y1,x2-x1);
                            angs.push_back(ang);

                            if(angs.rows > 1)
                                cout<<"neighbor point: ";
                            else
                                cout<<"self point: ";
                            cout<<"("<<x1<<" "<<y1<<") -> ("<<x2<<" "<<y2<<") with angle "<<ang<<endl;
                        }
                    }
                }

                cout<<"angs: "<<angs<<endl;

                // Compute mean and stddev
                Scalar mean, stddev;
                meanStdDev(angs,mean,stddev);
                Debug(0)<<"mean and stddev: "<<mean[0]<<" "<<stddev[0]<<endl;

                if(isOutlier(angs.at<double>(0,0), mean[0], stddev[0], angs.rows))
                    Debug(0)<<"outlier"<<endl;

                if(angs.rows-1 < minneighbors || isOutlier(angs.at<double>(0,0), mean[0], stddev[0], angs.rows)) {
                    matchList.first.erase(matchList.first.begin()+i);
                    matchList.second.erase(matchList.second.begin()+i);
                    Debug(0)<<"removed"<<endl;
                }
            }
            */
        }
    }


    void visualizeMatches(IplImage *img1, IplImage *img2, PointMatchList &pointList, string title)
    {
        if(img1!=NULL && img2!=NULL) {

            IplImage *stacked = Hess::stack_imgs( img1, img2 );
            CvPoint pt1, pt2;
            for(unsigned int i=0; i<pointList.first.size(); i++) {
                pt1 = pointList.first[i];
                pt2 = pointList.second[i];
                pt2.y += img1->height;
                //if(i==32 || i==48 || i==41)
                //    cvLine( stacked, pt1, pt2, CV_RGB(0,0,255), 1, 8, 0 );
                //else
                cvLine( stacked, pt1, pt2, CV_RGB(255,0,255), 1, 8, 0 );
            }
            Hess::display_big_img( stacked, (char*)title.c_str() );
        }
    }


    bool isOutlier(double x, double mean, double stddev, int numPoints)
    {
        // CHAUVENET'S CRITERION
        double stddevdist = fabs(x-mean);
        double prob = 1.0-erf((stddevdist/stddev)/sqrt(2));
        if(prob*numPoints < 0.3) //0.5
            return true;
        return false;
    }


} //end of namespace


