#ifndef IMAGEUTILS_H
#define IMAGEUTILS_H

#include "defs.h"
#include <limits>
#include <vector>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/calib3d/calib3d.hpp>

using namespace std;
using namespace cv;


namespace iu {

    //! Eliminate invalid feature matches using several modules
    //! @filterlevel defines the modules to active (default: use all modules)
    //! 1st bit: use fundemantal matrix
    //! 2nd bit: remove outliers based on distance
    //! 3rd bit: use spatial consistency to remove invalid matches
    int enforceFundMatConstraint(Mat &F, PointMatchList &pointList, int filterlevel=0b111, int img2Height=0, IplImage *img1=NULL, IplImage *img2=NULL);

    Mat calcFundamentalMat(PointMatchList *pointList, int method=0);
    Mat calcFundamentalMat_Ransac(PointMatchList *pointList, int numiter, int method=0, int numsamples=8, int mininliers=10);

    //! Fundamental matrix error computation using the error function selected by the errMethod
    //! @errMethod=0 use resudial error
    //! @errMethod=1 use sampsons error
    //! @errThreshold error threshold
    double calcFundMatError(Mat &F, PointMatchList &pointList, int errMethod=0, double errThreshold=-1);
    double calcFundMatError_Resudial(Mat &F, PointMatchList &pointList, double errThreshhold=1.5);
    double calcFundMatError_Sampsons(Mat &F, PointMatchList &pointList, double errThreshhold=0.1);

    //! Removes duplicate point match entries
    //! @TODO: These duplicates come from Hess' code (don't know why, maybe due to integer typecasting?)
    void removeDuplicates(PointMatchList &pointList);

    // Removes outliers based on distances between matching feature pairs
    void removeOutliers(PointMatchList &pointList);

    // Returns the number of minimum unique features in feature match list
    int getUniqueFeats(PointMatchList &pointList);

    // eliminate feature matches that do not comply with the neighbors' transformation
    // @nn is the radius of the neighborhood
    void applySpatialConsistency(PointMatchList &pointList, int imgHeight, int nn=5);

    Mat normalizeLine(Mat &L);
    void forceRank2(Mat &F);

    void visualizeMatches(IplImage *img1, IplImage *img2, PointMatchList &pointList, string title);

    //! Check if a number is outlier using CHAUVENET'S criterion
    bool isOutlier(double x, double mean, double stddev, int numPoints);
}

#endif // IMAGEUTILS_H
