#include <QMenu>
#include <QMessageBox>
#include <QPalette>
#include <QtGui>
#include "imagewidget.h"
#include <sstream>


#include "sift.h"
#include "imgfeatures.h"
#include "utils.h"
#include "kdtree.h"
#include "xform.h"



ImageWidget::ImageWidget(QWidget *parent)
    : QWidget(parent)
{
    image = NULL;

    imageLabel = new QLabel(this);
    imageLabel->setSizePolicy(QSizePolicy::Ignored,QSizePolicy::Ignored);

    // If we omitted to set the imageLabel's scaledContents property,
    // zooming in would enlarge the QLabel, but leave the pixmap at
    // its original size, exposing the QLabel's background.
    imageLabel->setScaledContents(true);

    imageDescription = new QLabel(this);
    imageDescription->setScaledContents(true);
    imageDescription->setVisible(false);
    QPalette textpalet;
    textpalet.setColor(QPalette::Foreground,Qt::black);
    imageDescription->setPalette(textpalet);

    createActions();

    setAttribute(Qt::WA_DeleteOnClose);
}

ImageWidget::~ImageWidget() {
    //cout<<"destructor for imagewidget"<<endl;

    if(image!=NULL) {
        delete image;
    }
    if(imageLabel!=NULL) {
        delete imageLabel;
    }
    if(imageDescription!=NULL) {
        delete imageDescription;
    }
    if(imageInfoAct!=NULL) {
        delete imageInfoAct;
    }
}

void ImageWidget::closeEvent(QCloseEvent *)
{
    //cout<<"close event for imagewidget"<<endl;
    delete this;
}


QSize ImageWidget::sizeHint () const {
    return imageLabel->pixmap()->size();
}

QSize ImageWidget::size () const {
    return this->sizeHint();
}


void ImageWidget::createActions() {
    imageInfoAct = new QAction(tr("Image &Info"), this);
    imageInfoAct->setShortcut(tr("Ctrl+I"));
    imageInfoAct->setEnabled(true);
    connect(imageInfoAct, SIGNAL(triggered()), this, SLOT(imageInfo()));

    showFeaturesAct = new QAction(tr("Show &Features"), this);
    showFeaturesAct->setShortcut(tr("Ctrl+F"));
    showFeaturesAct->setEnabled(true);
    connect(showFeaturesAct, SIGNAL(triggered()), this, SLOT(showFeatures()));
}

void ImageWidget::imageInfo() {
    QMessageBox::about(this, tr("Image Info"),imageDescription->text());
}


void ImageWidget::contextMenuEvent(QContextMenuEvent *event) {
    QMenu contextMenu(this);
    contextMenu.addAction(imageInfoAct);
    contextMenu.addAction(showFeaturesAct);
    contextMenu.exec(event->globalPos()+QPoint(3,3));
}


bool ImageWidget::openImage(QString filename) {

    QPixmap pix(filename);

    IplImage *imgrgb,*imggray;
    imgrgb=cvLoadImage(filename.toStdString().c_str());
    imggray = cvCreateImage(cvSize(imgrgb->width,imgrgb->height),IPL_DEPTH_8U,1);

    // apply gaussian with a small window size to remove some high freq noise
    int gaussian = 3;
    cvCvtColor(imgrgb,imggray,CV_RGB2GRAY);
    cvSmooth(imggray,imggray,CV_GAUSSIAN,gaussian,gaussian);
    cvReleaseImage(&imgrgb);

    // extract sift features
    struct Hess::feature *feats;
    int nPts = Hess::sift_features(imggray,&feats);

    QPainter painter1(&pix);
    QPen Red((QColor::QColor(255,255,0)),2);
    painter1.setPen(Red);

    /*
    int crossWidth = 6;
    for(int i=0; i<nPts; i++) {
        int x = feats[i].x;
        int y = feats[i].y;
        painter1.drawLine(x-crossWidth/2,y-crossWidth/2,x+crossWidth/2,y+crossWidth/2);
        painter1.drawLine(x-crossWidth/2,y+crossWidth/2,x+crossWidth/2,y-crossWidth/2);
    }
    */

    imageLabel->setPixmap(pix);
    imageLabel->show();


    return true;
/*
    QPixmap pix;
    stringstream descr;

    QString filecheck = filename.toLower();
    if(filecheck.endsWith(".pgm")) {
        if(image==NULL)
            image = new Image();
        image->loadImage(filename.toStdString(),Image::PGM);
        if(image->getImageFormat()==Image::GRAY) {
            QImage qimage((uchar *) image->getImageData(), image->getWidth(), image->getHeight(),QImage::Format_Indexed8);
            QVector<QRgb> colorTable;
            for(int i=0;i<256;i++)
                colorTable.push_back(QColor(i,i,i).rgb());

            qimage.setColorTable(colorTable);
            pix.convertFromImage(qimage);
            descr << "Image format: " << image->getImageFormat() << endl;
        }
    }
    else if(filecheck.endsWith(".ppm")) {
        if(image==NULL)
            image = new Image();
        image->loadImage(filename.toStdString(),Image::PPM);
        if(image->getImageFormat()==Image::RGB) {
            QImage qimage((uchar *) image->getImageData(), image->getWidth(), image->getHeight(), QImage::Format_RGB888);
            pix.convertFromImage(qimage);
            descr << "Image format: " << image->getImageFormat() << endl;
        }
    }
    else {
        return false;
    }

    if(image->numFeatures() < 0)
        image->extractFeatures();

    imageLabel->setPixmap(pix);
    imageLabel->show();


    descr << "Filename: " << filename.toStdString();
    descr << "\nSize: " << pix.width() << " x " << pix.height();
    descr << "\nNumFeatures: " << image->numFeatures();
    string description = descr.str();
    imageDescription->setText(QString(description.c_str()));
    return true;
*/
}


bool ImageWidget::openImage(Image *img) {
    if(img==NULL) {
        string description = "ERROR: NULL image is loaded\n";
        imageDescription->setText(QString(description.c_str()));
        imageDescription->show();
        return false;
    }

    if(image!=NULL) {
        delete image;
        image = NULL;
    }

    image = new Image(img);

    if(image->getImageData()==NULL) {
        image->loadImageFromFile();
    }

    QPixmap pix;
    if(image->getImageFormat()==Image::RGB) {
        QImage qimage((uchar *) image->getImageData(), image->getWidth(), image->getHeight(), QImage::Format_RGB888);
        pix.convertFromImage(qimage);
    }
    else if(image->getImageFormat()==Image::GRAY) {
        QImage qimage((uchar *) image->getImageData(), image->getWidth(), image->getHeight(),QImage::Format_Indexed8);
        QVector<QRgb> colorTable;
        for(int i=0;i<256;i++)
            colorTable.push_back(QColor(i,i,i).rgb());

        qimage.setColorTable(colorTable);
        pix.convertFromImage(qimage);
    }

    imageLabel->setPixmap(pix);
    imageLabel->show();

    if(image->numFeatures() < 0)
        image->extractFeatures();

    stringstream descr;
    descr << "Filename: " << image->getFilename();
    descr << "\nImage format: " << image->getImageFormat();
    descr << "\nSize: " << image->getWidth() << " x " << image->getHeight();
    descr << "\nNumFeatures: " << image->numFeatures();
    descr << "\nCoord: " << image->getPose().x<<" "<<image->getPose().y;
    string description = descr.str();
    imageDescription->setText(QString(description.c_str()));
    return true;
}


/*
bool ImageWidget::saveImage(QString filename, Image::FileFormat fileformat) {
    return false;
}
*/

void ImageWidget::showFeatures() {
    image->extractFeatures();

    QPixmap pix(*imageLabel->pixmap());
    QPainter p(&pix);
    QPen pen;
    pen.setColor(QColor("yellow")); // feature color
    pen.setWidth(1);
    p.setPen(pen);
    int crossSize = 4; // size of the crosses representing extracted feature locations
    for(int i=0; i < image->numFeatures(); i++) {
        Feat *f = image->features[i];
        SiftProp *sp = (SiftProp*)f->prop;
        double x = sp->loc.X();
        double y = sp->loc.Y();
        p.drawLine(x-crossSize/2,y-crossSize/2,x+crossSize/2,y+crossSize/2);
        p.drawLine(x+crossSize/2,y-crossSize/2,x-crossSize/2,y+crossSize/2);
    }
    p.end();
    imageLabel->setPixmap(pix);
    imageLabel->show();
}
