/***************************************************************
 *
 * Written by: Gorkem Erinc, gerinc@gmail.com
 *
 * Description:
 *    Creates the graph structure within an appearance-based
 *    map and handles all graph related functions.
 *
 ***************************************************************/

#include "simgraph.h"
#include "imageutils.h"

#include <stdlib.h>
#include <string>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <ctime>
#include <gvc.h>

#include <boost/graph/graphviz.hpp>
#include <boost/graph/connected_components.hpp>
#include <boost/graph/bc_clustering.hpp>
#include <boost/graph/iteration_macros.hpp>
#include <boost/algorithm/string.hpp>


using namespace std;
using namespace boost;
using namespace iu;
using namespace converter;

// colors taken from web color schema 13
string colors[150] = {"deepskyblue","green","violetred1","yellow","purple","red1","orange1","lawngreen","brown","blue1","lime","aliceblue","antiquewhite","aqua","aquamarine","azure","beige","bisque","black","blanchedalmond","blue","blueviolet","brown","burlywood","cadetblue","chartreuse","chocolate","coral","cornflowerblue","cornsilk","crimson","cyan","darkblue","darkcyan","darkgoldenrod","darkgray","darkgreen","darkgrey","darkkhaki","darkmagenta","darkolivegreen","darkorange","darkorchid","darkred","darksalmon","darkseagreen","darkslateblue","darkslategray","darkslategrey","darkturquoise","darkviolet","deeppink","dimgray","dimgrey","dodgerblue","firebrick","floralwhite","forestgreen","fuchsia","gainsboro","ghostwhite","gold","goldenrod","gray","grey","greenyellow","honeydew","hotpink","indianred","indigo","ivory","khaki","lavender","lavenderblush","lemonchiffon","lightblue","lightcoral","lightcyan","lightgoldenrodyellow","lightgray","lightgreen","lightgrey","lightpink","lightsalmon","lightseagreen","lightskyblue","lightslategray","lightslategrey","lightsteelblue","lightyellow","limegreen","linen","magenta","maroon","mediumaquamarine","mediumblue","mediumorchid","mediumpurple","mediumseagreen","mediumslateblue","mediumspringgreen","mediumturquoise","mediumvioletred","midnightblue","mintcream","mistyrose","moccasin","navajowhite","navy","oldlace","olive","olivedrab","orange","orangered","orchid","palegoldenrod","palegreen","paleturquoise","palevioletred","papayawhip","peachpuff","peru","pink","plum","powderblue","red","rosybrown","royalblue","saddlebrown","salmon","sandybrown","seagreen","seashell","sienna","silver","skyblue","slateblue","slategray","slategrey","snow","springgreen","steelblue","tan","teal","thistle","tomato","turquoise","violet","wheat","yellowgreen"};


SimGraph::SimGraph()
{
    numConnComp = -1;
    minDegree = -1;
    maxDegree = -1;
    avgDegree = -1;

    gWeight = get(edge_weight,g);
    gEdge = get(edgeptr_t(),g);

    // threshold to label an edge valid or invalid (for visualization)
    falseDistThreshdold = 300; // in pixels

    stats.clear();
}

SimGraph::~SimGraph()
{
    cout<<"simgraph destructor"<<endl;
    g.clear(); // is this even necessary?
    pair<edgeIterator, edgeIterator> erange = edges(g);
    edgeIterator e;
    for(e=erange.first; e!=erange.second; e++) {
        cvReleaseMat(&(gEdge[*e].F));
    }
}


void SimGraph::setDataDir(string dir)
{
    dataDir = dir;
}




// Given an image pointer create a vertex
vertexDescriptor SimGraph::addVertex(Image *img) {
    vertexDescriptor v = add_vertex(g);
    g[v].index = img->getIndex(); // set vertexindex = imageindex
    g[v].image = img; // set vertexindex = imageindex
    if(num_vertices(g)==1)
        g[v].clusterId = 0;
    else
        g[v].clusterId = -g[v].index;
    char idx[10];
    sprintf(idx,"%d",g[v].index);
    g[v].label = string(idx);

    lastVertex = &(g[v]);

    img->vertex = v; //set/update image's ve                                                                               rtex pointer
    vertexLookup[g[v].index] = v;

    cv::Point2d center(img->getPose().x,img->getPose().y);
    g[v].gridmapPose.center = center;
    g[v].gridmapPose.theta = img->getPose().z;
    return v;
}



bool SimGraph::removeVertex(int vIndex)
{
    //cout<<"deleting vertex: "<<vIndex<<endl;
    if(vertexLookup.count(vIndex)) {

        vertexDescriptor v = vertexLookup[vIndex];

        // delete image
        delete g[v].image;
        g[v].image = NULL;

        // delete from the invertedIndex & OccuranceList

        // delete from vertexLookup
        vertexLookup.erase(vIndex);

        // delete from g
        clear_vertex(v,g);
        remove_vertex(v,g);

        rebuildVertexLookup();

        //cout<<"deleted"<<endl;
        return true;
    }
    return false;
}

bool SimGraph::removeEdge(int v1Index, int v2Index)
{
    if(vertexLookup.count(v1Index) && vertexLookup.count(v2Index)) {
        cout<<"removing edge between "<<v1Index<<" "<<v2Index<<endl;

        Edge *e = getEdge(v1Index,v2Index);
        if(e!=NULL) {

            vertexDescriptor v1 = vertexLookup[v1Index];
            vertexDescriptor v2 = vertexLookup[v2Index];
            remove_edge(v1,v2,g);
            remove_edge(v2,v1,g);
            return true;
        }
    }
    return false;
}


bool SimGraph::removeEdges(vector<pair<int,int> > lim1, vector<pair<int,int> > lim2) {
    bool ret = false;
    vector<pair<int,int> > myedges;
    pair<edgeIterator, edgeIterator> erange = edges(g);
    for(edgeIterator e=erange.first; e!=erange.second; e++) {
        int v1 = g[source(*e,g)].index;
        int v2 = g[target(*e,g)].index;
        myedges.push_back(make_pair(v1,v2));
    }
    for(int k=0; k<myedges.size(); k++) {
        int v1 = myedges[k].first;
        int v2 = myedges[k].second;
        //cout<<"considering: "<<v1<<" "<<v2<<endl;

        for(int i=0; i<lim1.size(); i++) {
            for(int j=0; j<lim2.size(); j++) {
                cout<<"limits: "<<lim1[i].first<<" "<<lim1[i].second<<" "<<lim2[i].first<<" "<<lim2[i].second<<endl;
                if((v1 >= lim1[i].first && v2 >= lim2[i].first && !(lim1[i].second>-1 && v1 > lim1[i].second) && !(lim2[i].second>-1 && v2 > lim2[i].second)) ||
                   (v2 >= lim1[i].first && v1 >= lim2[i].first && !(lim1[i].second>-1 && v2 > lim1[i].second) && !(lim2[i].second>-1 && v1 > lim2[i].second))) {
                    removeEdge(v1,v2);
                    ret = true;
                }
            }
        }
    }
    //cout<<"done deleting"<<endl;
    return ret;
}


// If two vertices belong to two different clusters,
// after adding an edge between two vertices update cluster IDs
edgeDescr SimGraph::addEdge(int v1Index, int v2Index) {
    if(vertexLookup.count(v1Index) && vertexLookup.count(v2Index)) {
        edgeDescr e = add_edge(vertexLookup[v1Index],vertexLookup[v2Index],g);

        if(e.second) {
            gEdge[e.first].v1 = v1Index;
            gEdge[e.first].v2 = v2Index;

            vertexDescriptor v1 = vertexLookup[v1Index];
            vertexDescriptor v2 = vertexLookup[v2Index];

            cv::Point2d center1 = g[v1].gridmapPose.center;
            cv::Point2d center2 = g[v2].gridmapPose.center;
            gEdge[e.first].gridmapPose.center = (center1+center2)*0.5;
            gEdge[e.first].gridmapPose.boxHeight = 20;
            gEdge[e.first].gridmapPose.boxWidth = 20;

            gWeight[e.first] = 1;

            // no proper clusterId is assigned
            if(g[v1].clusterId==-g[v1].index) {
                g[v1].clusterId = g[v2].clusterId;
            }
            else if(g[v1].clusterId != g[v2].clusterId) {
                computeConnectedComponents();
            }
            return e;
        }
    }
    else {
        edgeDescriptor e;
        return make_pair(e,false);
    }
}


// Given a vertexDescriptor and all feature matches, create necessary edges between vertices
bool SimGraph::addEdges(vertexDescriptor v, CandidateList &approvedCandidates, ImageTable &imageLookup, ofstream *outstr) {

    bool isAdded = false;

    if(approvedCandidates.size()>0) {

        // Create maximum maxVertexDegree edges for this vertex, starting from the one with most number of feature matches
        int edgecount = 0;
        int candcount = 0;
        bool cont = true;
        CandidateList::const_iterator it=approvedCandidates.end();
        int candlimit = approvedCandidates.size();
        do{
            it--;
            candcount++;
            int imgIndex = g[v].image->getIndex();
            int candimgIndex = it->first.second;
            int numofsupporters = it->first.first; // sorted by
            CvMat* F = it->second.first;
            PointMatchList* inliers = it->second.second;

            vertexDescriptor v2 = imageLookup[candimgIndex]->vertex;

            int vdegree = degree(v2,g);
            //cout<<"v:"<<g[v2].image->getIndex()<<" "<<degree(v2,g)<<" "<<in_degree(v2,g)<<" "<<out_degree(v2,g)<<endl;


            if(vdegree >= appConf.MAXVERTEXDEGREE) {
                // target vertex has too many edges already
                // delete one to open space for the new edge

                Debug(1)<<"WARNING: TARGET HAS TOO MANY EDGES BUT FOR THE MOMENT I DO NOT REMOVE ONE OF THEM BEFORE I INSERT THIS"<<endl;
                /*
                Debug(1)<<"Target vertex has too many edges. Removing one of them randomly (uniform)"<<endl;

                // WARNING: Since the graph is undirectional, we are only considering out edges
                // (not in&out edges --they are the same for undirectional graphs)

                int numedges = out_degree(v2,g);
                uniform_int<> distribution(0,numedges-1);
                variate_generator<boost::mt19937&,boost::uniform_int<> > myrandom(randengine,distribution);

                int chosenvertex = myrandom();
                //cout<<"selected vertex to be deleted:"<<chosenvertex<<endl;

                pair<outedgeIterator, outedgeIterator> erange = out_edges(v2,g);
                outedgeIterator e = erange.first;
                for(int i=0;i<chosenvertex;i++) {
                    e++;
                }

                Debug(1)<<"removing edge between vertices: "<<g[source(*e,g)].index<<" and "<<g[target(*e,g)].index<<endl;

                remove_edge(*e,g);
*/
            }

            //cout<<"edge btw: "<<g[v].index<<" & "<<g[v2].index<<endl;

            edgeDescr e = add_edge(v,v2,g);
            if(e.second) {
                // printout added edge info
                if(outstr!=NULL) {
                    timespec ts;
                    clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&ts);
                    double time = ts.tv_sec + (ts.tv_nsec)/1000000000.0;
                    (*outstr)<<fixed<<time<<"\t"<<numofsupporters<<endl;
                }


                edgecount++;
                gEdge[e.first].numMatches = numofsupporters;
                gEdge[e.first].weight = numofsupporters;
                gEdge[e.first].F = F;
                gEdge[e.first].matches = inliers;
                gEdge[e.first].v1 = g[v].index;
                gEdge[e.first].v2 = g[v2].index;

                cv::Point2d center1 = g[v].gridmapPose.center;
                cv::Point2d center2 = g[v2].gridmapPose.center;
                gEdge[e.first].gridmapPose.center = (center1+center2)*0.5;
                gEdge[e.first].gridmapPose.boxHeight = 20;
                gEdge[e.first].gridmapPose.boxWidth = 20;


                double dist = norm(Mat(center1),Mat(center2));
                if(dist > falseDistThreshdold) {
                    gEdge[e.first].customVar = -1;
                    gEdge[e.first].cost = (std::numeric_limits<double>::max)();
                    //cout<<"FALSE EDGE"<<endl;
                }
                else {
                    gEdge[e.first].cost = 1/(double)numofsupporters;
                    gEdge[e.first].customVar = 0;
                }
                //gWeight[e.first] = gEdge[e.first].cost;
                gWeight[e.first] = 1;


                // no proper clusterId is assigned
                if(g[v].clusterId==-g[v].index) {
                    g[v].clusterId = g[v2].clusterId;
                }
                else if(g[v].clusterId != g[v2].clusterId) {
                    //int c1 = min(g[v].clusterId,g[v2].clusterId);
                    //int c2 = max(g[v].clusterId,g[v2].clusterId);
                    //connectivityMap[make_pair(c1,c2)] = true;
                    computeConnectedComponents();
                }

                Debug(2)<<"Edge created between images "<<imgIndex<<" and "<<candimgIndex<<" with "<<numofsupporters<<" valid feature matches"<<endl;
                isAdded = true;
            }
            else {
                printf("ERROR: Edge insertion between images %d and %d failed\n",imgIndex,candimgIndex);
                //delete F;
                cvReleaseMat(&F);
                delete inliers;
            }

            if(candcount == candlimit) { // all candidates are checked
                cont = false;
            }

        }while(cont);
    }
    return isAdded;
}


void SimGraph::rebuildVertexLookup() {
    vertexLookup.clear();
    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    for(vertexIterator v=vrange.first; v!=vrange.second; v++) {
        vertexLookup[g[*v].index] = *v;
    }
}

Vertex* SimGraph::getVertex(int index) {
    if(vertexLookup.count(index)){
        return &g[vertexLookup[index]];
    }
    else
        return NULL;
}

Vertex* SimGraph::getVertex(adjacencyIterator ai) {
    return &g[*ai];
}

Edge* SimGraph::getEdge(int index1,int index2) {
    if(vertexLookup.count(index1) && vertexLookup.count(index2)) {
        pair<edgeDescriptor,bool> edgebool = edge(vertexLookup[index1],vertexLookup[index2],g);
        if(edgebool.second)
            return &gEdge[edgebool.first];
        else {
            edgebool = edge(vertexLookup[index2],vertexLookup[index1],g);
            if(edgebool.second)
                return &gEdge[edgebool.first];
        }
    }
    return NULL;
}

pair<adjacencyIterator, adjacencyIterator> SimGraph::getEdges(int index) {
    return adjacent_vertices(vertexLookup[index], g);
}






/****************************************************/
/********************** GRAPHVIZ ********************/
/****************************************************/


int SimGraph::printGraph(string graphfile, string imagefile, string imageformat, string graphlayout)
{    
    Debug(2)<<"Print graph"<<endl;
    if(imageformat=="png" || imageformat=="gif" || imageformat=="jpg" || imageformat=="jpeg"
       || imageformat=="eps" || imageformat=="pdf" || imageformat=="ps" || imageformat=="dot" || imageformat=="plain")
    {
        char systemcall[200];
        sprintf(systemcall,"%s -T%s %s -o %s", graphlayout.c_str(), imageformat.c_str(), graphfile.c_str(), imagefile.c_str());
        ::system(systemcall);
        return 1;
    }
    else {
        cout << "ERROR: Cannot print the graph. Unknown file format" <<endl;
        return 0;
    }
}



void SimGraph::computeGraphLayoutCirc(int &w, int &h, double r, double dist) {
    int N = numVertices();
    double ang = 2.0*M_PI/(double)N;
    double R = dist/(2.0*sin(ang));
    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    int i=0;
    cv::Point2d center(R+2*r,R+2*r);
    w = center.x*2;
    h = center.y*2;
    for(vertexIterator v=vrange.first; v!=vrange.second; v++) {
        double a = (-M_PI/2.0)+(i*ang);
        double x = R*cos(a);
        double y = R*sin(a);
        g[*v].circPose.center = cv::Point2d(x,y)+center;
        g[*v].circPose.boxWidth = 2*r;
        g[*v].circPose.boxHeight = 2*r;
        i++;
    }
    pair<edgeIterator, edgeIterator> erange = edges(g);
    for(edgeIterator e=erange.first; e!=erange.second; e++) {
        cv::Point2d p1 = g[source(*e,g)].circPose.center;
        cv::Point2d p2 = g[target(*e,g)].circPose.center;
        gEdge[*e].circPose.center = (p1+p2)*0.5;
        gEdge[*e].circPose.boxWidth = 2*r;
        gEdge[*e].circPose.boxHeight = 2*r;
    }
}


void SimGraph::computeGraphLayout(string filename, string layout)
{
    Debug(2)<<"Compute graph layout"<<endl;
    GVC_t *_gvc;
    graph_t *_g;

    if(filename=="") {
        filename = dataDir+layout+".graph";
        writeGraph(filename);
    }
    FILE *fp = fopen(filename.c_str(), "r");

    _gvc = gvContext(); // this has to be called first
    _g = agread(fp);
    gvLayout(_gvc, _g, layout.c_str());
    gvRender(_gvc, _g, "dot", NULL);

    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    vertexIterator v;
    for(v=vrange.first; v!=vrange.second; v++) {
        string indx = int2str(g[*v].index);
        Agnode_t* node = agfindnode(_g,&indx[0]);

        char* s;
        s = agget(node,"pos");
        string position(s);
        size_t found1;
        found1 = position.find(",");
        string x(position, 0,found1);
        string y(position, found1+1);

        // these are in inches... 1 inch = 72 points
        s = agget(node,"width");
        string width(s);
        s = agget(node,"height");
        string height(s);

        // slight modifications: don't know the underlying reason of the mismatch between coords and image points
        //       BoundingBox ap(str2double(x)+5,str2double(y)+3,72*str2double(width),72*str2double(height));
        //       vertexDispTable[g[*v].index] = ap;

        g[*v].graphPose.center = cv::Point2d(str2double(x)+5,str2double(y)+3);
        g[*v].graphPose.boxWidth = str2double(width)*72;
        g[*v].graphPose.boxHeight = str2double(height)*72;
    }


    pair<edgeIterator, edgeIterator> erange = edges(g);
    edgeIterator e;
    for(e=erange.first; e!=erange.second; e++) {
        string indx1 = int2str(g[source(*e,g)].index);
        Agnode_t* node1 = agfindnode(_g,&indx1[0]);
        string indx2 = int2str(g[target(*e,g)].index);
        Agnode_t* node2 = agfindnode(_g,&indx2[0]);
        Agedge_t *edg = agfindedge(_g, node1, node2);

        pointf pos = edg->u.label->pos;
        int x = pos.x;
        int y = pos.y;

        pointf dimen = edg->u.label->dimen;
        int width = dimen.x;
        int height = dimen.y;

        gEdge[*e].graphPose.center = cv::Point2d(x+5,y+3);
        gEdge[*e].graphPose.boxWidth = width;
        gEdge[*e].graphPose.boxHeight = height;
    }

    gvFreeLayout(_gvc, _g);
    agclose(_g);
    gvFreeContext(_gvc);
}






void SimGraph::writeGraph(string filename)
{
    Debug(2)<<"Writing graph to file: "<<filename<<endl;

    ofstream out(filename.c_str());

    out << "graph G {" << endl;
    out << "dpi=72;"<<endl;
    out << "node [";
    out << "style=filled";
    out << ", color=deepskyblue";
    out << ", fillcolor=deepskyblue";
    out << ", width=\"0.75\"";
    out << ", height=\"0.60\"";
    out << "];"<< endl;

    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    vertexIterator v;
    for(v=vrange.first; v!=vrange.second; v++) {
        out << g[*v].index      << " [";
        out << "vertexIndex=\"" << g[*v].index              << "\"";
        out << ", imageIndex=\""  << g[*v].image->getIndex() << "\"";
        out << ", label=\""  << g[*v].label << "\"";
        out << ", color=\""  << colors[g[*v].mapId] << "\"";
        out << ", fillcolor=\""  << colors[g[*v].mapId] << "\"";
        out << ", mapId=\""  << g[*v].mapId << "\"";
        out << ", clusterId=\""  << g[*v].clusterId << "\"";
        out << ", customVar=\""  << g[*v].customVar << "\"";
        out << "];\n";
    }

    out << "edge [";
    out << "color=azure3";
    out << ", fontsize=12";
    out << ", fontcolor=gray21";
    out << ", arrowhead=none";
    out << "];" << endl;
    pair<edgeIterator, edgeIterator> erange = edges(g);
    edgeIterator e;
    for(e=erange.first; e!=erange.second; e++) {
        int indx1 = g[source(*e,g)].index;
        int indx2 = g[target(*e,g)].index;
        out << indx1 << "--";
        out << indx2 << " [";
        out << "label=\"" << gEdge[*e].weight << "\"";
        out << ", numMatches=\"" << gEdge[*e].numMatches << "\"";
        out << ", edgeweight=\"" << gEdge[*e].weight << "\"";
        out << ", cost=\"" << gEdge[*e].cost << "\"";
        out << ", mapId=\"" << gEdge[*e].mapId << "\"";
        out << ", customVar=\"" << gEdge[*e].customVar << "\"";
        out << ", F=\"";
        if(gEdge[*e].F!=NULL) {
            for(int i=0; i < gEdge[*e].F->rows; i++) {
                for(int j=0; j < gEdge[*e].F->cols; j++) {
                    out << cvmGet(gEdge[*e].F,i,j) << " ";
                }
            }
        }
        else {
            for(int i=0; i < 9; i++) {
                out << 0.0 << " ";
            }
        }
        out << "\"";

        //int nummatches = gEdge[*e].matches->first.size();
        out << ", matches=\"";
        for(int i=0; i < gEdge[*e].numMatches; i++) {
            out << "(" << (gEdge[*e].matches->first)[i].x << "," << (gEdge[*e].matches->first)[i].y << ")-(" << (gEdge[*e].matches->second)[i].x << "," << (gEdge[*e].matches->second)[i].y << ") ";
        }

        out << "\"";
        out << "];\n";
    }
    out << "}\n";
    out.close();
}


void SimGraph::readGraph(string filename, ImageTable &imageLookup)
{
    Debug(2)<<"Reading graph from file: "<<filename<<endl;

    ifstream ifs(filename.c_str());
    string line,temp;
    istringstream parse;
    getline(ifs,line); // graph g {
    getline(ifs,line); // dpi=72;
    while(getline(ifs,line)) {

        if(line[0]!='}') { // last line
            size_t found = line.find("--");
            if(found==string::npos) { // vertex
                parse.clear();
                parse.str(line);
                parse >> temp;
                if (temp.compare("node")==0 || temp.compare("edge")==0)
                    continue;

                size_t found1, found2;

                int vertexIndex = -1;
                found1 = line.find("vertexIndex=\"");
                if(found1!=string::npos) {
                    found1+=13;
                    found2 = line.find("\"",found1);
                    string vindex(line, found1,found2-found1);
                    vertexIndex= atoi(vindex.c_str());
                }

                int imageIndex = -1;
                found1 = line.find("imageIndex=\"");
                if(found1!=string::npos) {
                    found1+=12;
                    found2 = line.find("\"",found1);
                    string iindex(line, found1,found2-found1);
                    imageIndex = atoi(iindex.c_str());
                }

                string vertexLabel = "";
                found1 = line.find("label=\"");
                if(found1!=string::npos) {
                    found1+=7;
                    found2 = line.find("\"",found1);
                    string vlabel(line, found1,found2-found1);
                    vertexLabel = vlabel;
                }
                if(vertexLabel=="") {
                    ostringstream myStream;
                    myStream << vertexIndex;
                    vertexLabel = myStream.str();
                }

                string color = "";
                found1 = line.find("color=\"");
                if(found1!=string::npos) {
                    found1+=7;
                    found2 = line.find("\"",found1);
                    string vlabel(line, found1,found2-found1);
                    color = vlabel;
                }
                int mapId = 0;
                if(color!="") {
                    for(unsigned int c=0; c<colors->size()-1; c++) {
                        if(color==colors[c])
                            mapId = c;
                    }
                }

                found1 = line.find("mapId=\"");
                if(found1!=string::npos) {
                    found1+=9;
                    found2 = line.find("\"",found1);
                    string gid(line, found1,found2-found1);
                    mapId = atoi(gid.c_str());
                }

                int clusterId = 0;
                found1 = line.find("clusterId=\"");
                if(found1!=string::npos) {
                    found1+=11;
                    found2 = line.find("\"",found1);
                    string cid(line, found1,found2-found1);
                    clusterId = atoi(cid.c_str());
                }

                int customVar = 0;
                found1 = line.find("customVar=\"");
                if(found1!=string::npos) {
                    found1+=11;
                    found2 = line.find("\"",found1);
                    string cvar(line, found1,found2-found1);
                    customVar = atoi(cvar.c_str());
                }


                vertexDescriptor v;
                v = add_vertex(g);

                g[v].index = vertexIndex;
                g[v].label = vertexLabel;
                g[v].mapId = mapId;
                g[v].clusterId = clusterId;
                g[v].customVar = customVar;
                Image *img = imageLookup[imageIndex];
                g[v].image = img; // creating link from vertex to image
                img->vertex = v;  // creating link from image to vertex
                vertexLookup[vertexIndex] = v;

                cv::Point2d center(img->getPose().x,img->getPose().y);
                g[v].gridmapPose.center = center;
                g[v].gridmapPose.theta = img->getPose().z;
            }
            else{ // edge
                size_t found1, found2, found3;
                found1 = line.find("--");
                found2 = line.find(" ",found1+2);
                string sv1(line, 0,found1);
                string sv2(line, found1+2,found2-found1-2);
                int v1index = atoi(sv1.c_str());
                int v2index = atoi(sv2.c_str());

                found1 = line.find("\"",found2+1);
                found2 = line.find("\"",found1+1);
                string label(line, found1+1,found2-found1-1);

                found1 = line.find("\"",found2+1);
                found2 = line.find("\"",found1+1);
                string nmatch(line, found1+1,found2-found1-1);
                int numMatches = atoi(nmatch.c_str());

                found1 = line.find("\"",found2+1);
                found2 = line.find("\"",found1+1);
                string w(line, found1+1,found2-found1-1);
                double weight;
                parse.clear();
                parse.str(w);
                parse >> weight;

                found1 = line.find("\"",found2+1);
                found2 = line.find("\"",found1+1);
                string c(line, found1+1,found2-found1-1);
                double cost;
                parse.clear();
                parse.str(c);
                parse >> cost;


                int mapId = 0;
                found1 = line.find("mapId=\"");
                if(found1!=string::npos) {
                    found1+=9;
                    found2 = line.find("\"",found1);
                    string gid(line, found1,found2-found1);
                    mapId = atoi(gid.c_str());
                }

                int customVar = 0;
                found1 = line.find("customVar=\"");
                if(found1!=string::npos) {
                    found1+=11;
                    found2 = line.find("\"",found1);
                    string cvar(line, found1,found2-found1);
                    customVar = atoi(cvar.c_str());
                }


                found1 = line.find("\"",found2+1);
                found2 = line.find("\"",found1+1);
                string fmat(line, found1+1,found2-found1-1);
                CvMat* F = cvCreateMat(3,3,CV_64FC1);
                double val;
                parse.clear();
                parse.str(fmat);
                for(int i=0; i < 3; i++) {
                    for(int j=0; j < 3; j++) {
                        parse >> val;
                        cvmSet(F,i,j,val);
                    }
                }

                PointMatchList *matches = new PointMatchList;

                found1 = line.find("\"",found2+1);
                found2 = line.find("\"",found1+1);
                string matchstr(line, found1+1,found2-found1-1);
                parse.clear();
                parse.str(matchstr);
                for(int i=0; i < numMatches; i++) {
                    parse >> temp;
                    found1 = temp.find("(");
                    found2 = temp.find("-",found1+1);
                    found3 = temp.find(")",found2+1);
                    string first(temp, found1+1,found2-found1-2);
                    string second(temp, found2+2,found3-found2-2);

                    found1 = first.find(",");
                    string x1(first,0,found1);
                    string y1(first,found1+1,first.length()-found1-1);                  
                    cv::Point2d p1(atof(x1.c_str()),atof(y1.c_str()));

                    found1 = second.find(",");
                    string x2(second,0,found1);
                    string y2(second,found1+1,second.length()-found1-1);
                    cv::Point2d p2(atof(x2.c_str()),atof(y2.c_str()));

                    matches->first.push_back(p1);
                    matches->second.push_back(p2);
                }

                edgeDescr e = add_edge(vertexLookup[v1index], vertexLookup[v2index] ,g);
                gEdge[e.first].numMatches = numMatches;
                gEdge[e.first].weight = weight;               
                gEdge[e.first].F = F;
                gEdge[e.first].matches = matches;
                gEdge[e.first].v1 = v1index;
                gEdge[e.first].v2 = v2index;
                gEdge[e.first].mapId = mapId;
                gEdge[e.first].customVar = customVar;
                gEdge[e.first].cost = 1/(double)numMatches;

                cv::Point2d center1 = g[vertexLookup[v1index]].gridmapPose.center;
                cv::Point2d center2 = g[vertexLookup[v2index]].gridmapPose.center;
                gEdge[e.first].gridmapPose.center = (center1+center2)*0.5 + cv::Point2d(0,0);

                gWeight[e.first] = 1;
            }
        }
    }
    ifs.close();
#if DEBUG > 1
    printf("\tThe graph has %d vertices and %d edges\n",num_vertices(g),num_edges(g));
#endif
}



void SimGraph::updateStats()
{
    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    vertexIterator v;
    minDegree = std::numeric_limits<int>::max(); // a big number
    for(v=vrange.first; v!=vrange.second; v++) {
        int vdegree = degree(*v,g);
        if(minDegree > vdegree)
            minDegree = vdegree;
        if(maxDegree < vdegree)
            maxDegree = vdegree;
    }
    avgDegree = (2.0*num_edges(g))/(double)num_vertices(g);

    vector<int> component(num_vertices(g));
    numConnComp = connected_components(g, &component[0]);

    graphStat s;
    s.minDegree = minDegree;
    s.maxDegree = maxDegree;
    s.avgDegree = avgDegree;
    s.numConnComp = numConnComp;
    s.numVertices = num_vertices(g);
    s.numEdges = num_edges(g);
    stats.push_back(s);

    /*
    vector<int>::size_type i;
    cout << "Total number of components: " << num << endl;
    for (i = 0; i != component.size(); ++i)
      cout << "Vertex " << i <<" is in component " << component[i] << endl;
    cout << endl;
    */
}

void SimGraph::writeStats(string filename)
{
    ofstream out(filename.c_str());
    if(out.is_open()) {
        out << "minDegree maxDegree avgDegree numConnectedComp numVertices numEdges" << endl;

        for(unsigned int i=0; i<stats.size();i++) {
            out << " " << stats[i].minDegree;
            out << " " << stats[i].maxDegree;
            out << " " << stats[i].avgDegree;
            out << " " << stats[i].numConnComp;
            out << " " << stats[i].numVertices;
            out << " " << stats[i].numEdges;
        }
        out << endl;
        out.close();
    }
    else {
        cout<<"File could not be opened to write for stats: "<<filename<<endl;
    }
}


// This function updates the vertex and disp data so that when mapwidget is clicked we know which vertex to show
void SimGraph::updateDisplay(DispMode mode)
{
    Debug(0)<<"Update Display"<<endl;
    vertexDispTable.clear();
    edgeDispTable.clear();

    // VERTICES
    Mat vdata(0,2,CV_64F);
    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    vertexIterator v;
    for(v=vrange.first; v!=vrange.second; v++) {
        vertexDispTable.push_back(&g[*v]);

        if(mode==GRAPHMODE) {
            Mat m1(g[*v].graphPose.center);
            m1 = m1.t();
            vdata.push_back(m1);
        }
        else if(mode==GRIDMAPMODE) {
            Mat m(1,2,CV_64F);
            m.at<double>(0,0) = g[*v].gridmapPose.center.x;
            m.at<double>(0,1) = g[*v].gridmapPose.center.y;
            vdata.push_back(m);
        }
        else if(mode==CIRCMODE) {
            Mat m1(g[*v].circPose.center);
            m1 = m1.t();
            vdata.push_back(m1);
        }
    }
    if(vdata.rows > 0) {
        idxGraphV.setClusters(vdata);
        idxGraphV.buildClusterIndex();
    }
    Debug(0)<<"DONE - VERTICES"<<endl;


    // EDGES
    Mat edata(0,2,CV_64F);
    pair<edgeIterator, edgeIterator> erange = edges(g);
    edgeIterator e;
    for(e=erange.first; e!=erange.second; e++) {
        edgeDispTable.push_back(&gEdge[*e]);

        if(mode==GRAPHMODE) {
            Mat m1(gEdge[*e].graphPose.center);
            m1 = m1.t();
            edata.push_back(m1);
        }
        else if(mode==GRIDMAPMODE) {
            Mat m1(gEdge[*e].gridmapPose.center);
            m1 = m1.t();
            edata.push_back(m1);
        }
        else if(mode==CIRCMODE) {
            Mat m1(gEdge[*e].circPose.center);
            m1 = m1.t();
            edata.push_back(m1);
        }
    }
    if(edata.rows > 0) {
        idxGraphE.setClusters(edata);
        idxGraphE.buildClusterIndex();
    }
    Debug(0)<<"DONE - EDGES"<<endl;
}


// Given the coordinates of the mouse click and distance, find the closest vertex
Vertex* SimGraph::getClosestVertex(int x, int y, double &distance)
{
    Mat q(1,2,CV_64F);
    q.at<double>(0,0) = x;
    q.at<double>(0,1) = y;

    int nn = 1;
    Mat indices(1,nn,CV_32S);
    Mat dists(1,nn,CV_32F);

    idxGraphV.getIndex()->knnSearch(q, indices, dists, nn, cv::flann::SearchParams(128));

    distance = dists.at<float>(0,0);
    int num = indices.at<int>(0,0);

    return vertexDispTable[num];
}


// Given the coordinates of the mouse click and distance, find the closest edge
Edge* SimGraph::getClosestEdge(int x, int y, double &distance)
{
    Mat q(1,2,CV_64F);
    q.at<double>(0,0) = x;
    q.at<double>(0,1) = y;

    int nn = 1;
    Mat indices(1,nn,CV_32S);
    Mat dists(1,nn,CV_32F);

    idxGraphE.getIndex()->knnSearch(q, indices, dists, nn, cv::flann::SearchParams(128));

    distance = dists.at<float>(0,0);
    int num = indices.at<int>(0,0);

    return edgeDispTable[num];
}



void SimGraph::resetVertexLabels()
{
    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    for(vertexIterator v=vrange.first; v!=vrange.second; v++) {
        ostringstream myStream;
        myStream << g[*v].index;
        g[*v].label = myStream.str();
        g[*v].mapId = 0;
    }
}



// applies clustering based on brandes_betweenness_centrality by removing edges...
int SimGraph::computeClusters(vector<int> &components, double max_centrality) {

    Graph gnew = g;

    if(max_centrality<0)
        max_centrality = (double)boost::num_vertices(gnew)/10.0;

    // std::map used for convenient initialization
    typedef std::map<edgeDescriptor, int> StdEdgeIndexMap;
    StdEdgeIndexMap my_e_index;
    // associative property map needed for iterator property map-wrapper
    typedef boost::associative_property_map< StdEdgeIndexMap > EdgeIndexMap;
    EdgeIndexMap e_index(my_e_index);

    // We use setS as edge-container -> no automatic indices
    // -> Create and set it explicitly
    int i = 0;
    BGL_FORALL_EDGES(edge, gnew, Graph)
    {
        my_e_index.insert(std::pair< edgeDescriptor, int >( edge, i));
        ++i;
    }

    // Define EdgeCentralityMap
    vector< double > e_centrality_vec(boost::num_edges(gnew), 0.0);
    // Create the external property map
    boost::iterator_property_map< std::vector< double >::iterator, EdgeIndexMap >
            e_centrality_map(e_centrality_vec.begin(), e_index);

    // Define VertexCentralityMap
    typedef boost::property_map< Graph, boost::vertex_index_t>::type VertexIndexMap;
    VertexIndexMap v_index = get(boost::vertex_index, gnew);
    vector< double > v_centrality_vec(boost::num_vertices(gnew), 0.0);
    // Create the external property map
    boost::iterator_property_map< std::vector< double >::iterator, VertexIndexMap >
            v_centrality_map(v_centrality_vec.begin(), v_index);

    // Calculate the vertex and edge centralites
    // Can be used to get an initial impression about the edge centrality values for the graph
    brandes_betweenness_centrality( gnew, v_centrality_map, e_centrality_map );

    // 'false' means here that no normalization is performed, so edge centrality-values can become big
    // If set to 'true', values will range between 0 and 1 but will be more difficult to use for this
    // illustrative example.
    boost::bc_clustering_threshold<double> terminate(max_centrality, gnew, false);

    // Do the clustering
    // Does also calculate the brandes_betweenness_centrality and stores it in e_centrality_map
    betweenness_centrality_clustering( gnew, terminate, e_centrality_map );


    ofstream graphStream("clusters.dot");
    boost::write_graphviz(graphStream, gnew);
    graphStream.close();
    ::system("dot -Tpng clusters.dot -o clusters.png");
    ::system("rm clusters.dot");


    // Compute cluster index for each vertex
    components.resize(num_vertices(gnew));
    int num = connected_components(gnew, &components[0]);
    /*
    cout<<"Number of connected components: "<<num<<endl;
    for(int i=0; i<component.size(); i++) {
        cout<<"Vertex "<<i<<" is in component "<<component[i]<<endl;
    }
    */
    return num;
}





// if the vertex with that index is connected to the last inserted vertex cluster
bool SimGraph::isConnected(CandidateList &candidates) {

    if(lastVertex==NULL)
        return false;

    BOOST_FOREACH(CandidateList::value_type &c, candidates) {
        int c1 = min(lastVertex->clusterId,getVertex(c.first.second)->clusterId);
        int c2 = max(lastVertex->clusterId,getVertex(c.first.second)->clusterId);

        if(c1 == c2)
            return true;
    }
    return false;
}


bool SimGraph::sameConnectedComponent(int v1Index, int v2Index) {

    if(getVertex(v1Index)->clusterId == getVertex(v2Index)->clusterId)
        return true;

    //if(connectivityMap[make_pair(c1,c2)]==true)
    //    return true;

    return false;
}


void SimGraph::computeConnectedComponents()
{
    //struct timeval tic,toc;
    //gettimeofday(&tic,NULL);

    vector<int> component(num_vertices(g));
    numConnComp = connected_components(g, &component[0]);

    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    int i=0;
    components.clear();
    for(vertexIterator v=vrange.first; v!=vrange.second; v++) {
        g[*v].clusterId = component[i];
        components[g[*v].clusterId].push_back(g[*v].index);
        i++;
    }

    //gettimeofday(&toc,NULL);
    //double elapsed = (toc.tv_sec -tic.tv_sec) + (toc.tv_usec - tic.tv_usec)/1000000.0;
    //cout<<"CONNCOMP TIME: "<<elapsed<<endl;

}


bool SimGraph::improvesGraph(CandidateList &candidates) {
    for(CandidateList::iterator it1=candidates.begin(); it1!=candidates.end(); it1++) {
        CandidateList::iterator ittemp=it1;
        for(CandidateList::iterator it2=++ittemp; it2!=candidates.end(); it2++) {

            vertexDescriptor v1 = vertexLookup[it1->first.second];
            vertexDescriptor v2 = vertexLookup[it2->first.second];

            Debug(0)<<"path between: "<<g[v1].index<<" "<<g[v2].index<<endl;

            if(g[v1].clusterId != g[v2].clusterId)
                return true;

            int d = graphDistance(it1->first.second,it2->first.second);

            if(d==-1) { // there is no connection between two vertices
                Debug(0)<<"NO PATH"<<endl;
                return true;
            }
            if(d-2-1 >= 8) { // net hop improvement is greater than or equal to 8
                Debug(0)<<"path found! size: "<<d<<endl;
                return true;
            }
        }
    }
    return false;
}

int SimGraph::graphDistance(int i1, int i2) {

    vertexDescriptor v1 = vertexLookup[i1];
    vertexDescriptor v2 = vertexLookup[i2];

    vector<vertexDescriptor> mypath;
    AStar mysearch;
    if(mysearch.searchThroughGraph(g,v1,v2,mypath)) {
        return mypath.size();
    }
    else { // there is no connection between two vertices
        return -1;
    }
}


int SimGraph::isPathShorterThan(int v1,int v2, int threshold) {

    //cout<<"Dist between: "<<v1<<" and "<<v2;

    if(v1==v2)
        return 0;

    pair<adjacencyIterator, adjacencyIterator> arange;
    map<int,bool> isProcessed;
    vector<int> nextRound;
    vector<int> nextRound2;
    nextRound.push_back(v1);

    for (int k=1; k<=threshold; k++) {
        while(nextRound.size()>0) {
            int v = nextRound.back();
            isProcessed[v] = true;
            nextRound.pop_back();
            arange = adjacent_vertices(vertexLookup[v], g);
            for(adjacencyIterator it=arange.first; it!=arange.second; it++) {
                int cand = g[*it].index;
                if(cand==v2) {
                    cout<<" "<<k<<endl;
                    return k;
                }
                if(isProcessed[cand]==false) {
                    nextRound2.push_back(cand);
                }
            }
        }
        for(int i=0;i<nextRound2.size();i++) {
            nextRound.push_back(nextRound2[i]);
        }
        nextRound2.clear();
    }
    //cout<<" larger than "<<threshold<<endl;
    return -1;
}


Mat SimGraph::graphLaplacian() {
    const int n = num_vertices(g);
    Mat L(n,n,CV_32F);

    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    int i=0;
    for(vertexIterator v1=vrange.first; v1!=vrange.second; v1++) {
        int j=0;
        for(vertexIterator v2=vrange.first; v2!=vrange.second; v2++) {
            if(i==j) {
                L.at<float>(i,j) = 0;
            }
            else {
                edgeDescr edgebool = edge(*v1,*v2,g);
                if(edgebool.second) {
                    if(gEdge[edgebool.first].customVar < 0 && false) {
                        L.at<float>(i,j) = 0.0;
                        L.at<float>(j,i) = 0.0;
                    }
                    else {
                        L.at<float>(i,j) = -1;
                        L.at<float>(j,i) = -1;
                    }
                }
                else {
                    L.at<float>(i,j) = 0;
                    L.at<float>(j,i) = 0;
                }
            }
            j++;
        }
        i++;
    }

    for(int i=0; i<L.rows; i++) {
        float sum=0.0;
        for(int j=0; j<L.cols; j++) {
            sum += L.at<float>(i,j);
        }
        L.at<float>(i,i) = -sum;
    }

/*
    // check if laplacian is computed correctly
    for(int i=0; i<n; i++) {
        float a = 0.0;
        for(int j=0; j<n; j++) {
            if(i!=j)
                a += L.at<float>(i,j);
        }
        //cout<<"compare L: "<<a<<" "<<L.at<float>(i,i)<<" for i:"<<i<<endl;
        if(-a!=L.at<float>(i,i)) {
            cout<<"ERROR!!!!!!!!!!!!!! in algComp"<<endl;
            exit(0);
        }
    }
*/

    return L;
}



void SimGraph::writeSimMatrix(string filename)
{
    Debug(2)<<"Writing SimMatrix to file: "<<filename<<endl;

    ofstream out(filename.c_str());

    pair<vertexIterator,vertexIterator> vrange = vertices(g);
    for(vertexIterator v1=vrange.first; v1!=vrange.second; v1++) {
        for(vertexIterator v2=vrange.first; v2!=vrange.second; v2++) {
            pair<edgeDescriptor, bool> res = edge(*v1,*v2, g);
            if(res.second)
                out<<1<<" ";
            else
                out<<0<<" ";
        }
        out<<endl;
    }

    out.close();
}

