#ifndef SIMGRAPH_H
#define SIMGRAPH_H

#include <boost/random.hpp>
#include "dictionary.h"
#include "astar.h"
#include "image.h"
#include "graph.h"



typedef boost::unordered_map<int,vertexDescriptor> VertexTable; // <vertexIndex,vertexDescriptor>

// This map structure is used since this way it keeps it sorted wrt numberOfMatches
typedef map<pair<int,int>,pair<CvMat*,PointMatchList*> > CandidateList; // <pair<numberOfMatches, machedImageIndex>

// @TODO: make sure I check if the cluster index is within the size of the color array
extern string colors[150]; // to visualize each cluster with different colors...


struct graphStat {
    int numConnComp;    // number of connected components
    int minDegree;      // minimum vertex degree
    int maxDegree;      // maximum vertex degree
    double avgDegree;   // average vertex degree
    int numEdges;
    int numVertices;
};


class SimGraph
{
    friend class SimMap;
public:
    enum DispMode {
        GRAPHMODE = 0,
        GRIDMAPMODE,
        CIRCMODE,
        MATRIXMODE
    };
private:
    Graph g;
    //AStar mysearch; // this slows the process very much.. @TODO: maybe I should provide this as an option
    VertexTable vertexLookup; // is a map of <vertexIndex, vertexDescriptor> to link vertices with images
    WeightList::type gWeight;
    EdgeList::type gEdge;

    double falseDistThreshdold;

    // DISP RELATED
    cvflannIndex::Index<double,double> idxGraphV;
    cvflannIndex::Index<double,double> idxGraphE;
    vector<Vertex*> vertexDispTable;
    vector<Edge*> edgeDispTable;


    string dataDir; // local path to temporary folder

    // Statistics for the graph: an element to the list will be added whenever a new vertex/edge is created
    int numConnComp;
    int minDegree;
    int maxDegree;
    double avgDegree;


public:
    SimGraph();
    ~SimGraph();

    int numVertices() { return num_vertices(g); }
    int numEdges() { return num_edges(g); }

    vertexDescriptor addVertex(Image *img);

    // add all possible edges for a vertex encoded in the candidate list
    bool addEdges(vertexDescriptor v, CandidateList &matches, ImageTable &imageLookup, ofstream *a=NULL);
    edgeDescr addEdge(int v1Index, int v2Index);

    bool removeVertex(int v);
    bool removeEdge(int v1,int v2);
    // remove edges whose IDs are between entered limits
    // @TODO: I need to explain the user the format of the input
    bool removeEdges(vector<pair<int,int> > lim1, vector<pair<int,int> > lim2);


    void rebuildVertexLookup();
    Vertex* getVertex(int index);
    Vertex* getVertex(adjacencyIterator ai);
    Edge* getEdge(int index1,int index2);
    pair<adjacencyIterator, adjacencyIterator> getEdges(int index);
    VertexTable* getVertexLookup() { return &vertexLookup; }
    int getVertexDegree(int index) { return degree(vertexLookup[index],g); }

    void writeGraph(string filename);
    void readGraph(string filename, ImageTable &imageLookup);
    void writeSimMatrix(string filename);

    void setDataDir(string);

    Mat graphLaplacian(); // compute graph laplacian

    // VISUALIZATION FUNCTIONS
    void computeGraphLayoutCirc(int &w, int &h, double r, double dist);
    void computeGraphLayout(string filename, string layout="dot");
    int printGraph(string graphfile, string imagefile, string imageformat, string graphlayout);

    vector<Vertex*>* getVertexPositions() { return &vertexDispTable; }
    vector<Edge*>* getEdgePositions() { return &edgeDispTable; }

    void resetVertexLabels();
    void updateDisplay(DispMode mode);
    Vertex* getClosestVertex(int x, int y, double &distance);
    Edge* getClosestEdge(int x, int y, double &distance);


    // CLUSTERING FUNCTIONS
    int computeClusters(vector<int> &clusters, double max_centrality=-1.0);
    map<int,vector<int> > components; // holds cluster/connectedComponent membership of all vertices
    Vertex *lastVertex; // last inserted vertex.. for quick similarity comparison
    bool isConnected(CandidateList &candidates); // if the query img is connected to the candidates
    bool sameConnectedComponent(int v1Index, int v2Index); // if two vertices are within the same conn component
    void computeConnectedComponents();
    bool improvesGraph(CandidateList &candidates); // if the insertion of the img will improve graph connectivity
    int isPathShorterThan(int v1,int v2, int threshold); // this is based on the number of hops
    int graphDistance(int v1,int v2);


    // STATISTICS
    void updateStats();
    void writeStats(string filename);
    // for stats.. not used currently
    vector<graphStat> stats;

    void test() {}
};

#endif // SIMGRAPH_H
