/***************************************************************
 *
 * Written by: Gorkem Erinc, gerinc@gmail.com
 *
 * Description:
 *    Creates appearance-based maps with following functions:
 *    map building, analysis, visualization, image insertion.
 *
 ***************************************************************/

#include "simmap.h"
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <QFileInfo>
#include <QStringList>
#include <QString>
#include <QDir>
#include <QRegExp>
#include "configuration.h"
#include <string>

#include <boost/graph/graphviz.hpp>

// For clustering
#include <boost/graph/bc_clustering.hpp>
#include <boost/graph/iteration_macros.hpp>
#include <boost/graph/connected_components.hpp>

#include <boost/graph/johnson_all_pairs_shortest.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>

#include <queue>
#include <time.h>

using namespace converter;
using namespace iu;
using namespace std;



SimMap::SimMap()
{
    mapId = 0;

    maxIndex = -1;

    // Check if sys folder exists..
    QDir dir;
    if(!dir.exists("sys"))
        dir.mkdir("sys");

    // Create a temp folder in sys
    QString dirpath("sys/map");
    int count = 0;
    while(true) {
        count++;
        dirpath = "sys/map" + QString::number(count);

        if(!dir.exists(dirpath)) {
            dir.mkdir(dirpath);
            break;
        }
    }
    dataDir = dirpath.toStdString()+"/";

    dict = new Dictionary(appConf.CLUSTER_FILE, appConf.INDEX_FILE);
    sg = new SimGraph();
    sg->setDataDir(dataDir);

    filepath = "";
}

// load the map stored in a file
SimMap::SimMap(path fname)
{
    mapId = 0;

    maxIndex = -1;

    // Check if sys folder exists..
    QDir dir;
    if(!dir.exists("sys"))
        dir.mkdir("sys");

    // Create a temp folder in sys
    QString dirpath("sys/map");
    int count = 0;
    while(true) {
        count++;
        dirpath = "sys/map" + QString::number(count);

        if(!dir.exists(dirpath)) {
            dir.mkdir(dirpath);
            break;
        }
    }
    dataDir = dirpath.toStdString()+"/";

    dict = NULL;
    sg = NULL;
    readMap(fname);
}



SimMap::~SimMap()
{
    delete dict;
    delete sg;
    cleanImageList();
    char systemcall[200];
    sprintf(systemcall,"rm -rf %s",dataDir.c_str());
    //cout<<"called: "<<systemcall<<endl;
    std::system(systemcall);
}



bool SimMap::isEmpty() {
    if(sg->numVertices() == 0 && imageLookup.size() == 0)
        return true;
    return false;
}

void SimMap::cleanImageList() {
    //cout<<"clear imagelist"<<endl;
    BOOST_FOREACH(ImageTable::value_type &it, imageLookup) {
        delete it.second;
    }
    imageLookup.clear();
}


// reserved for testing
void SimMap::test() {
}

// reserved for testing
void SimMap::test(QString imagefile1,QString imagefile2) {

    cout<<"TEST"<<endl;
    /*
    Image *img1=new Image();
    Image *img2=new Image();
    img1->loadImage(imagefile1.toStdString(),Image::PPM);
    img2->loadImage(imagefile2.toStdString(),Image::PPM);
    img1->extractFeatures();
    img2->extractFeatures();

    vertexDescriptor v1 = addImageAsVertex(img1);
    //vertexDescriptor v2 = addImageAsVertex(img2);

    PointMatchMap matches;
    int maxvote = getVotes(img2, matches);
    CandidateList candidates;

    struct timeval tStart,tEnd;
    gettimeofday(&tStart,NULL);
    enforceGeometricConstraints(matches, candidates, true);
    gettimeofday(&tEnd,NULL);
    double elapsed = (tEnd.tv_sec -tStart.tv_sec) + (tEnd.tv_usec - tStart.tv_usec)/1000000.0;

    cout<<maxvote<<endl;
    cout<<candidates.size()<<endl;
    cout<<"time: "<<elapsed<<endl;
    */

    /*
    vector<int> clusters;
    int num = sg->computeClusters(clusters);
    cout<<num<<endl;
    */
}




int SimMap::chooseIndex(int i)
{
    int idx;

    // Choose a good index for the image
    if(i > maxIndex) {
        // index not in use
        maxIndex = i;
        idx = i;
    }
    else
        idx = ++maxIndex;

    Debug(1)<<"Chosen index: "<<idx<<endl;
    return idx;
}



int SimMap::getVotes(Image *img, PointMatchMap &matches) {
    int maxvote = -1;
    int localizedimage = -1;

    dict->findMatches(img, matches); // will be filled with PointMatches like <img1,img2>
    Debug(1)<<"total matches: "<<matches.size()<<endl;

    // Get the image with maximum votes
    BOOST_FOREACH(PointMatchMap::value_type &mm, matches) {
        PointMatchList *ml = &(mm.second);
        if((int)(ml->first.size()) > maxvote) {
            maxvote = (int)(ml->first.size());
            localizedimage = mm.first;
        }
    }

    Debug(1)<<"Max vote for image("<<localizedimage<<"): "<<maxvote<<endl;

    return maxvote;
}


bool SimMap::enforceGeometricConstraints(PointMatchMap &matches, CandidateList &candidates, bool forceSimilarImages)
{
    candidates.clear();

    bool isTooSimilar = false;

    // Check for lower bound
    BOOST_FOREACH(PointMatchMap::value_type &m, matches) {

        int candimgIndex = m.first;
        PointMatchList *pml = &m.second;

        //PointMatchList *pml = &m.second; //pml points to pair<pointList,pointList>
        if((int)pml->first.size() >= appConf.MINMATCHES) {

            Debug(1)<<"Candidate image match "<<candimgIndex<<" with "<<pml->first.size()<<" matches"<<endl;

            // Given two sets of properties
            // compute F and count number of supporting matches by enforcing minimal error
            // this will return the most probable F and the number of supporters for this F
            CvMat* F = cvCreateMat(3,3,CV_64FC1);
            cv::Mat t(F);

            int numofsupporters = enforceFundMatConstraint(t, *pml, 0b001);

            //int numofsupporters = pml->first.size();
            // if more than limit, quit process since there is AT LEAST one vertex too similar to this image
            if(numofsupporters > appConf.MAXMATCHES)
                isTooSimilar = true;

            if(isTooSimilar && forceSimilarImages==false) {
                Debug(0)<<"Too similar so quit looking"<<endl;
                return isTooSimilar;
            }
            else if(numofsupporters >= appConf.MINMATCHES) {
                PointMatchList *inliers = new PointMatchList;
                for(unsigned int k=0; k<pml->first.size(); k++) {
                    inliers->first.push_back(pml->first[k]);
                    inliers->second.push_back(pml->second[k]);
                }
                candidates[make_pair(numofsupporters,candimgIndex)] = make_pair(F,inliers);

                Debug(1)<<"passed with: "<<numofsupporters<<endl;
                //cout<<" PASSED"<<endl;
            }
            else {
                Debug(1)<<"failed with: "<<numofsupporters<<endl;
                cvReleaseMat(&F);
                //cout<<" FAILED"<<endl;
            }
        }
    }
    return isTooSimilar;
}




vector<int> SimMap::addImage(QString imagefile, int indx, bool force, int trackingMode, bool addEdge)
{
    Image *img = new Image();

    if(indx > -1) {
        img->setIndex(indx);
    }
    else { // try to extract index info from filename
        QFileInfo fi(imagefile);
        QRegExp reg("_0*(\\d+)\\.");
        reg.setPatternSyntax(QRegExp::RegExp2);
        if(reg.indexIn(fi.fileName()) != -1) {
            QString num = reg.cap(1);
            img->setIndex(num.toInt());
        }
    }
    Debug(2)<<"\n\nAttempting to add image: "<<imagefile.toStdString()<<endl;



    if(imagefile.endsWith(".pgm",Qt::CaseInsensitive)) {
        img->loadImage(imagefile.toStdString(),Image::PGM);
        return addImage(img,force,trackingMode,addEdge);
    }
    else if(imagefile.endsWith(".ppm",Qt::CaseInsensitive)) {
        img->loadImage(imagefile.toStdString(),Image::PPM);
        img->extractFeatures();

        return addImage(img,force,trackingMode,addEdge);
    }
    else {
        delete img;
        img = NULL;
        vector<int> temp;
        return temp;
    }
}


vertexDescriptor SimMap::addImageAsVertex(Image *img, bool keepData) {
    int oldIndex = img->getIndex();
    int chosenIndex = chooseIndex(img->getIndex());
    img->setIndex(chosenIndex);

    Debug(1)<<"index set to: "<<img->getIndex()<<" from: "<<oldIndex<<endl;

    // add the image to the list
    imageLookup[img->getIndex()] = img;

    // Create a vertex with that image
    vertexDescriptor v = sg->addVertex(img);
    getGraph()->g[v].oldindex = oldIndex;

    //sg->updateStats();

    // Add the location of the features from the image to the database
    // no ties with corresponding image pointer
    dict->addImage(img);

    if(!keepData)
        img->deleteData();

    return v;
}


vector<int> SimMap::addImage(Image *img, bool force, int trackingMode, bool addEdge) {

    Debug(2)<<"Processing image: "<<img->getFilename()<<endl;

    // Extract its features if not extracted before
    if(img->features.size()==0) {
        //cout<<"features extracted"<<endl;
        img->extractFeatures();
    }

    if(img->features.size()==0) {
        delete img;
        img = NULL;
        vector<int> temp;
        return temp;
    }

    vector<int> newIndices;

    // Find similarities between query image and all other images in the DB
    PointMatchMap matches;

    int maxvote = getVotes(img, matches);

    CandidateList candidates;

    if(force) {
        enforceGeometricConstraints(matches, candidates, true);

        vertexDescriptor v = addImageAsVertex(img);
        if(addEdge)
            sg->addEdges(v, candidates, imageLookup);
        newIndices.push_back(getGraph()->g[v].index);

        return newIndices;
    }


    bool backtracking = false;

    bool tooSimilar = enforceGeometricConstraints(matches, candidates, true);
    bool notEnough = (candidates.size()==0);

    if(notEnough) {
        Debug(0)<<"NOT ENOUGH - Start BT"<<endl;
        backtracking = true;
    }
    else {
        bool isconnected = getGraph()->isConnected(candidates);

        if(tooSimilar) {
            Debug(0)<<"TOO SIMILAR"<<endl;
            if(isconnected) { // has last cluster edge
                if(getGraph()->improvesGraph(candidates)) {
                    Debug(0)<<"CONNECTED - ADD"<<endl;
                    Debug(0)<<"Have edges"<<endl;
                    vertexDescriptor v = addImageAsVertex(img);
                    if(addEdge)
                        getGraph()->addEdges(v, candidates, imageLookup);
                    newIndices.push_back(getGraph()->g[v].index);

                    for(unsigned int i=0; i<trackingList.size(); i++) {
                        delete trackingList[i];
                    }
                    trackingList.clear();
                }
                else {
                    Debug(0)<<"CONNECTED - IGNORE"<<endl;
                    trackingList.push_back(img);
                }
            }
            else {
                Debug(0)<<"NOT CONNECTED - Start BT"<<endl;
                backtracking = true;
            }
        }
        else { // similar enough
            Debug(0)<<"SIMILAR ENOUGH"<<endl;
            if(isconnected) { // has last cluster edge
                Debug(0)<<"CONNECTED - ADD"<<endl;
                Debug(0)<<"Have edges"<<endl;
                vertexDescriptor v = addImageAsVertex(img);
                if(addEdge)
                    getGraph()->addEdges(v, candidates, imageLookup);
                newIndices.push_back(getGraph()->g[v].index);

                for(unsigned int i=0; i<trackingList.size(); i++) {
                    delete trackingList[i];
                }
                trackingList.clear();
            }
            else {
                Debug(0)<<"NOT CONNECTED - Start BT"<<endl;
                backtracking = true;
            }
        }
    }


    // Backtracking allows to go back to previously processed images that are ignored due to high similarity
    // This mechanism preserves the connectivity..
    if(backtracking) {
        Debug(0)<<"STARTING BT"<<endl;
        if(trackingList.size()==0) {
            Debug(0)<<"EMPTY TRACKINGLIST"<<endl;
            vertexDescriptor v = addImageAsVertex(img);
            newIndices.push_back(getGraph()->g[v].index);
        }
        else {
            vector<Image*> processList;
            processList.push_back(img);

            while(trackingList.size()>0) {
                Image* myimage = trackingList.back();
                trackingList.pop_back();

                cout<<"backtracking process image: "<<myimage->getFilename()<<endl;

                // Find similarities between query image and all other images in the DB
                PointMatchMap matches;
                getVotes(myimage, matches);

                CandidateList candidates;
                enforceGeometricConstraints(matches, candidates,true);

                bool isconnected = getGraph()->isConnected(candidates);
                if(isconnected)
                    Debug(0)<<"ISCONNECTED"<<endl;
                else
                    Debug(0)<<"NOT CONNECTED"<<endl;

                if (candidates.size() > 0 && isconnected) {   // at least one edge creation needed
                    Debug(0)<<"ADDING VERTEX"<<endl;
                    vertexDescriptor v = addImageAsVertex(myimage);
                    if(addEdge)
                        sg->addEdges(v, candidates, imageLookup);
                    newIndices.push_back(getGraph()->g[v].index);

                    for(unsigned int i=0; i<trackingList.size(); i++) {
                        delete trackingList[i];
                    }
                    trackingList.clear();
                }
                else { // no edges
                    Debug(0)<<"NO EDGES"<<endl;
                    if(trackingList.size()==0) {
                        Debug(0)<<"FAILED BACKTRACKING"<<endl;
                        // failed backtracking
                        vertexDescriptor v = addImageAsVertex(myimage);
                        newIndices.push_back(getGraph()->g[v].index);
                    }
                    else {
                        processList.push_back(myimage);
                    }
                }
            }

            for(unsigned int i=0; i<processList.size(); i++) {
                vector<int> ilist = addImage(processList[i]);
                for(unsigned int j=0; j<ilist.size(); j++) {
                    newIndices.push_back(ilist[j]);
                }
            }
        }
    }

    return newIndices;
}





// Given an image localize that image in the map
// and return a pointer to the localized image
int SimMap::localizeImage(Image *img) {
    int imgIndex = img->getIndex();

    PointMatchMap matches;
    dict->findMatches(img,matches);

    int maxvote = -1;
    int localizedImage = -1;

    BOOST_FOREACH(PointMatchMap::value_type &m, matches) {
        if((int)m.second.first.size() > appConf.MINMATCHES) {

            int candimgIndex = m.first;

            Debug(0)<<"Candidate image match ("<<imgIndex<<"-"<<candimgIndex<<") : # "<<m.second.first.size()<<" matches"<<endl;

            // Given two sets of properties
            // compute F and count number of supporting matches by enforcing minimal error
            // this will return the most probable F and the number of supporters for this F
            CvMat* F = cvCreateMat(3,3,CV_64FC1);
            PointMatchList *pl = &m.second;
            cv::Mat t(F);
            int numofsupporters = enforceFundMatConstraint(t, *pl);

            if(numofsupporters >= appConf.MINMATCHES) {
                if(maxvote < numofsupporters) {
                    maxvote = numofsupporters;
                    localizedImage = candimgIndex;
                }
            }
        }
    }

    // Currently we say the image is localized if it has more than "minVotes" votes
    if(localizedImage != -1) {
        printf("Image(%d) is localized to image(%d) with %d number of votes",img->getIndex(),localizedImage,maxvote);
    }
    else {
        Debug(0)<<"Image("<<img->getIndex()<<") could not be localized."<<endl;
    }
    return localizedImage;
}


bool SimMap::addEdge(int v1, int v2) {
    edgeDescr e = getGraph()->addEdge(v1,v2);
    return e.second;
}

bool SimMap::removeVertex(int vIndex) {

    Debug(0)<<"deleting vertex: "<<vIndex<<endl;
    if(getGraph()->vertexLookup.count(vIndex)) {
        // delete from imagelookup
        Vertex* v = getGraph()->getVertex(vIndex);
        imageLookup.erase(v->image->getIndex());

        InvIndex &inv = *getDictionary()->getInvIndex();

        for(int i=0; i<v->image->words.size(); i++) {
            inv[v->image->words[i].first].erase(v->image->getIndex());
            if(inv[v->image->words[i].first].size()==0) {
                inv.erase(v->image->words[i].first);
            }
        }

        bool result = getGraph()->removeVertex(vIndex);

        return result;
    }
    return false;
}

bool SimMap::removeEdge(int v1, int v2) {
    return getGraph()->removeEdge(v1,v2);
}

bool SimMap::removeEdges(vector<pair<int,int> > lim1, vector<pair<int,int> > lim2) {
    return getGraph()->removeEdges(lim1,lim2);
}

bool SimMap::removeConnectedComp(int c) {
    bool ret = false;

    getGraph()->computeConnectedComponents();

    typedef map<int,vector<int> > temptype;
    BOOST_FOREACH(temptype::value_type &t, getGraph()->components) {
        if(t.second.size() <= c) {
            // delete all vertices in this component
            for(int i=0; i<t.second.size(); i++) {
                int vIndex = t.second[i];
                Debug(0)<<"removing clusterVertex: "<<vIndex<<endl;
                removeVertex(vIndex);
                ret = true;
            }
        }
    }
    if(ret)
        getGraph()->computeConnectedComponents();
    getGraph()->updateStats();
    return ret;

}



bool SimMap::writeMap(path fpath, bool writeIndvfiles) {

    cout<<"Writing map into file ("<<fpath.string()<<")..."<<endl;

    // Create a temp folder in sys
    QString dirpath("sys/temp");
    QDir dir;
    int count = 0;
    while(true) {
        count++;
        dirpath = "sys/temp" + QString::number(count);

        if(!dir.exists(dirpath)) {
            dir.mkdir(dirpath);
            break;
        }
    }
    string tempfoldername = dirpath.toStdString() + "/";

    string filename = fpath.stem();

    string graphFilename = filename+".sg";
    string imagelistFilename = filename+".il";
    string dictFilename = filename+".di";
    string configFilename = "map.conf";

    dict->writeDictionary(tempfoldername+dictFilename);
    writeImageList(tempfoldername+imagelistFilename);
    sg->writeGraph(tempfoldername+graphFilename);

    // Write config file
    Debug(1)<<"Writing configuration to file: "<<tempfoldername+configFilename<<endl;

    ofstream conf((tempfoldername+configFilename).c_str());
    if(conf.is_open()) {
        conf << "files: " << endl;
        conf << "graphFilename = " << graphFilename << endl;
        conf << "dictFilename = " << dictFilename << endl;
        conf << "imagelistFilename = " << imagelistFilename << endl;
        conf << endl;
        conf << "configurations: " << endl;
        conf << "FMATSAMPLES = " << appConf.FMATSAMPLES << endl;
        conf << "FMATMININLIERS = " << appConf.FMATMININLIERS << endl;
        conf << "MAXMATCHES = " << appConf.MAXMATCHES << endl;
        conf << "MINMATCHES = " << appConf.MINMATCHES << endl;
        conf << "FEATUREDIM = " << appConf.FEATUREDIM << endl;
        conf << "CLUSTER_FILE = " << appConf.CLUSTER_FILE << endl;
        conf << "INDEX_FILE = " << appConf.INDEX_FILE << endl;
        conf << "TRACKINGMODE = " << appConf.TRACKINGMODE << endl;
        conf.close();
    }
    else {
        cout<<"ERROR: Configuration file ("<<configFilename<<") could not be opened to write"<<endl;
        return false;
    }

    char systemcall[200];

    // Compress these files to create a map file
    sprintf(systemcall,"cd %s && tar -cf %s * && cd ../..",tempfoldername.c_str(),fpath.string().c_str());
    ::system(systemcall);


    if(writeIndvfiles) {
        sprintf(systemcall,"mv %s%s .",tempfoldername.c_str(),graphFilename.c_str());
        ::system(systemcall);
        sprintf(systemcall,"mv %s%s .",tempfoldername.c_str(),imagelistFilename.c_str());
        ::system(systemcall);
        sprintf(systemcall,"mv %s%s .",tempfoldername.c_str(),dictFilename.c_str());
        ::system(systemcall);
        sprintf(systemcall,"mv %s%s .",tempfoldername.c_str(),configFilename.c_str());
        ::system(systemcall);
    }

    sprintf(systemcall,"rm -rf %s",tempfoldername.c_str());
    ::system(systemcall);

    setFilepath(fpath);

    return true;
}


bool SimMap::readMap(path fpath)
{
    //boost::filesystem::path filepath(fp.toStdString());
    cout<<"Reading map from file ("<<fpath.string()<<")..."<<endl;

    // First deallocate current data in the map
    if(sg!=NULL)
        delete sg;
    if(dict!=NULL)
        delete dict;
    cleanImageList();
    char systemcall[200];
    sprintf(systemcall,"rm -f %s*",dataDir.c_str());
    ::system(systemcall);

    // Extract these files to read the map file
    sprintf(systemcall,"tar -C %s -xf %s",dataDir.c_str(), fpath.string().c_str());
    ::system(systemcall);

    string configFilename = "map.conf";
    string graphFilename, imagelistFilename, dictFilename;

    // Read config file
    Debug(0)<<"Reading configuration to file: "<<dataDir+configFilename<<endl;

    ifstream conf((dataDir+configFilename).c_str());

    // To be able to read old map files where configuration file is saved as filename.conf instead of map.conf
    // This is changed because in the case of map filename change, it was not possible to open the map
    if(!conf.is_open()) {
        string filename = fpath.stem();
        configFilename = filename+".conf";
        conf.open((dataDir+configFilename).c_str());
    }

    if(conf.is_open()) {
        istringstream parse;
        string line, temp, valuestr;
        int value;
        while(getline(conf,line)) {
            parse.clear();
            parse.str(line);
            parse >> temp;

            if(temp=="graphFilename") {
                parse >> temp >> graphFilename;
            }
            else if(temp=="dictFilename") {
                parse >> temp >> dictFilename;
            }
            else if(temp=="imagelistFilename") {
                parse >> temp >> imagelistFilename;
            }
            else if(temp=="FMATSAMPLES") {
                parse >> temp >> value;
                appConf.FMATSAMPLES = value;
            }
            else if(temp=="FMATMININLIERS") {
                parse >> temp >> value;
                appConf.FMATMININLIERS = value;
            }
            else if(temp=="MAXMATCHES") {
                parse >> temp >> value;
                appConf.MAXMATCHES = value;
            }
            else if(temp=="MINMATCHES") {
                parse >> temp >> value;
                appConf.MINMATCHES = value;
            }
            else if(temp=="FEATUREDIM") {
                parse >> temp >> value;
                appConf.FEATUREDIM = value;
            }
            else if(temp=="CLUSTER_FILE") {
                parse >> temp >> valuestr;
                appConf.CLUSTER_FILE = valuestr;
            }
            else if(temp=="INDEX_FILE") {
                parse >> temp >> valuestr;
                appConf.INDEX_FILE = valuestr;
            }
            else if(temp=="TRACKINGMODE") {
                parse >> temp >> value;
                appConf.TRACKINGMODE = value;
            }
            else {
                continue;
            }
        }
        conf.close();
    }
    else {
        cout<<"ERROR: Configuration file ("<<configFilename<<") could not be opened to read"<<endl;
        return false;
    }

    dict = new Dictionary(appConf.CLUSTER_FILE,appConf.INDEX_FILE);
    sg = new SimGraph();
    sg->setDataDir(dataDir);

    readImageList(dataDir+imagelistFilename);
    dict->readDictionary(dataDir+dictFilename,imageLookup);
    sg->readGraph(dataDir+graphFilename,imageLookup);
    updateStats();



    // Check if map files contains all image files
    bool filecheck = true;
    BOOST_FOREACH(ImageTable::value_type &it, imageLookup) {
        Image * img = it.second;
        if(img->getFilename() == "") {
            filecheck = false;
            break;
        }
        else {
            //QFile f((dataDir+img->getFilename()).c_str());
            QFile f(img->getFilename().c_str());
            if (!f.exists()) {
                filecheck = false;
                break;
            }
        }
    }
    if (!filecheck) {
        cout<<"ERROR: Images cannot be loaded from the map file because the map file does not contain all images"<<endl;
        return false;
    }

    trackingList.clear();

    setFilepath(fpath);

    return true;
}


bool SimMap::writeImageList(string filename) {
    Debug(2)<<"Writing imageList to file: "<<filename<<endl;

    ofstream out(filename.c_str());
    if(!out.is_open())
        return false;
    BOOST_FOREACH(ImageTable::value_type &it, imageLookup) {
        out << "ImageIndex=" << it.second->getIndex() << ", ";
        //cout<<"write filename: "<<it.second->getFilename()<<endl;
        out << "Filename=" << it.second->getFilename() << ", ";
        out << "Fileformat=" << it.second->getFileFormat() << ", ";
        out << "Timestamp=" << it.second->getTimestamp().tv_sec << " " <<it.second->getTimestamp().tv_usec << ", ";
        out << "Location=" << it.second->getPose().x << " " << it.second->getPose().y << " " << it.second->getPose().z;
        out << endl;
    }
    out.close();
    return true;
}

bool SimMap::readImageList(string filename) {
    Debug(2)<<"Reading imageList from file: "<<filename<<endl;

    ifstream in(filename.c_str());
    if(!in.is_open())
        return false;
    string line;
    while(getline(in,line)) {
        if(!converter::isEmpty(line)) {
            size_t found1, found2;
            found1 = line.find("=");
            found2 = line.find(",",found1+1);
            string index(line,found1+1, found2-found1-1);
            int imageIndex = converter::str2int(index);

            found1 = line.find("=",found2+1);
            found2 = line.find(",",found1+1);
            string filename(line,found1+1, found2-found1-1);

            found1 = line.find("=",found2+1);
            found2 = line.find(",",found1+1);
            string fileformat(line,found1+1, found2-found1-1);

            found1 = line.find("=",found2+1);
            found2 = line.find(",",found1+1);
            string mytime(line,found1+1, found2-found1-1);

            found1 = line.find("=",found2+1);
            found2 = line.find(",",found1+1);
            string myloc(line,found1+1, found2-found1-1);

            found1 = line.find("=",found2+1);
            string words = line.substr(found1+1);

            Image* img = new Image();
            img->setIndex(imageIndex);
            if(imageIndex>maxIndex)
                maxIndex = imageIndex;

            if(!converter::isEmpty(filename))
                img->setFilename(filename);
            //cout<<"read filename: "<<filename<<endl;

            if(!converter::isEmpty(fileformat))
                img->setFileFormat(str2int(fileformat));
            if(!converter::isEmpty(mytime)) {
                stringstream parse(mytime);
                long int sec,usec;
                parse >> sec >> usec;
                struct timeval tv;
                tv.tv_sec = sec;
                tv.tv_usec = usec;
                img->setTimestamp(tv);
            }
            if(!converter::isEmpty(myloc)) {
                stringstream parse(myloc);
                double x, y, theta;
                parse >> x >> y >> theta;
                img->setPose(Pose(x,y,theta));
            }
            imageLookup[imageIndex] = img;
            img->deleteData();
        }
    }
    in.close();
    return true;
}


bool SimMap::saveImageToFile(Image *img, string path) {

    if(img->isEmpty()) {
        if(img->loadImageFromFile()==-1) {
            return false;
        }
    }

    char imgfilename[50];
    if(img->getImageFormat()==Image::RGB) {
        sprintf(imgfilename,"RGB_%05d.ppm",img->getIndex());
        img->saveImage(path+imgfilename, Image::PPM);
        img->setFilename(imgfilename);
        return true;
    }
    else if(img->getImageFormat()==Image::GRAY) {
        sprintf(imgfilename,"GRAY_%05d.pgm",img->getIndex());
        img->saveImage(path+imgfilename, Image::PGM);
        img->setFilename(imgfilename);
        return true;
    }
    return false;
}

void SimMap::printGraph(QString &imagefile, QString imageformat, QString graphlayout) {
    QDir dir(dataDir.c_str());
    if(imagefile.isEmpty()) {
        imagefile = dir.absolutePath()+"/"+graphlayout+"."+imageformat;
    }
    QString graphfile = dir.absolutePath()+"/"+graphlayout+".graph";
    sg->writeGraph(graphfile.toStdString());
    if(!sg->printGraph(graphfile.toStdString(), imagefile.toStdString(),imageformat.toStdString(),graphlayout.toStdString()))
        filepath = "";
    else
        sg->computeGraphLayout(graphfile.toStdString(), graphlayout.toStdString());
}


void SimMap::writeStats(QString filepath) {
    sg->writeStats(filepath.toStdString());
}




double SimMap::scoreMap()
{
    Mat d = getGraph()->graphLaplacian();
    int n = d.rows;

    //timespec tic,toc;
    //clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&tic);
    Mat E;
    eigen(d,E);
    //clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&toc);
    //double t = toc.tv_sec -tic.tv_sec + (toc.tv_nsec-tic.tv_nsec)/1000000000.0;
    //cout<<t<<endl;


    double s = E.at<float>(n-2,0);

    if(n>1)
        return s;
    else
        return 0;
}



/***********************************/
/********** TEMPORARY **************/
/***********************************/

// dump graph data to file in order to find the minimal algCon gain for addition of 1 edge
void SimMap::dumpGraph(SimMap *m1, SimMap *m2,map<int,int> &oldnewIndex)
{
    // ******************** TEMP START *************************
    const int n = num_vertices(getGraph()->g);

    // Create d matrix and a list of all possible edges
    Mat d(n,n,CV_32F);
    pair<vertexIterator,vertexIterator> vrange = vertices(getGraph()->g);
    int i=0;
    for(vertexIterator v1=vrange.first; v1!=vrange.second; v1++) {
        int j=0;
        for(vertexIterator v2=vrange.first; v2!=vrange.second; v2++) {
            if(i==j) {
                //d.at<float>(i,j) = degree(*v1,getGraph()->g); // not really! :)
                //since I am removing edges I should also remove them during degree computations
                pair<outedgeIterator, outedgeIterator> outrange = out_edges(*v1,getGraph()->g);
                pair<inedgeIterator, inedgeIterator> inrange = in_edges(*v1,getGraph()->g);
                outedgeIterator it1;
                inedgeIterator it2;

                int deg = 0;
                for(it1=outrange.first; it1!=outrange.second; it1++) {
                    if(getGraph()->gEdge[*it1].customVar>=0)
                        deg++;
                }
                for(it2=inrange.first; it2!=inrange.second; it2++) {
                    if(getGraph()->gEdge[*it2].customVar>=0)
                        deg++;
                }
                d.at<float>(i,j) = deg/2;
            }
            else {
                edgeDescr edgebool = edge(*v1,*v2,getGraph()->g);
                if(edgebool.second) {
                    if(getGraph()->gEdge[edgebool.first].customVar < 0)
                        d.at<float>(i,j) = 0;
                    else
                        d.at<float>(i,j) = -1;
                }
                else
                    d.at<float>(i,j) = 0;
            }
            j++;
        }
        i++;
    }


    // Create a list of all possible edges
    map<pair<int,int>,bool> edgelist;
    //randomConnect(m1, m2, oldnewIndex, MergeMethod::BRUTEORDER, MergeStopCriteria::PERCENTAGE, 200);
    updateStats();
    vrange = vertices(getGraph()->g);
    i=0;
    for(vertexIterator v1=vrange.first; v1!=vrange.second; v1++) {
        int j=0;
        for(vertexIterator v2=vrange.first; v2!=vrange.second; v2++) {
            if(i!=j) {
                if(getGraph()->g[*v1].mapId != getGraph()->g[*v2].mapId) {
                    edgeDescr edgebool = edge(*v1,*v2,getGraph()->g);
                    if(edgebool.second) {
                        if(getGraph()->gEdge[edgebool.first].customVar >= 0) {
                            edgelist[make_pair(min(i,j),max(i,j))] = true;
                        }
                        else
                            cout<<"false edge"<<endl;
                    }
                }
            }
            j++;
        }
        i++;
    }


    ofstream out1("_d.txt");
    out1<<n<<" "<<n<<endl;
    for(int i=0; i<n; i++) {
        for(int j=0; j<n; j++) {
            out1<<d.at<float>(i,j)<<" ";
        }
        out1<<endl;
    }
    out1.close();

    ofstream out2("_e.txt");
    map<pair<int,int>,bool>::iterator it;
    for(it=edgelist.begin(); it!=edgelist.end(); it++) {
        if(it->second) {
            out2<<it->first.first<<" "<<it->first.second<<endl;
        }
    }
    out2.close();;
}


void SimMap::dumpGraphHeader(ofstream &out, SimMap *m1, SimMap *m2) {
    out<<"m1: "<<m1->getGraph()->numVertices()<<endl;
    out<<"m2: "<<m2->getGraph()->numVertices()<<endl;
    out<<"-----------------------"<<endl;

    out<<"positions (x,y):"<<endl;
    pair<vertexIterator,vertexIterator> vrange1 = vertices(m1->getGraph()->g);
    pair<vertexIterator,vertexIterator> vrange2 = vertices(m2->getGraph()->g);

    for(vertexIterator v1=vrange1.first; v1!=vrange1.second; v1++) {
        out<<m1->getGraph()->g[*v1].gridmapPose.center.x<<" "<<m1->getGraph()->g[*v1].gridmapPose.center.y<<endl;
    }
    for(vertexIterator v2=vrange2.first; v2!=vrange2.second; v2++) {
        out<<m2->getGraph()->g[*v2].gridmapPose.center.x<<" "<<m2->getGraph()->g[*v2].gridmapPose.center.y<<endl;
    }
}


