#ifndef SIMMAP_H
#define SIMMAP_H

#include "simgraph.h"
#include "image.h"
#include <QString>
#include <QStringList>
#include "time.h"
#include <boost/filesystem/path.hpp>
#include <boost/random.hpp>

using namespace boost::filesystem;


class SimMap
{
    int mapId;

private:
    SimGraph *sg;
    Dictionary *dict;
    ImageTable imageLookup;     // is a map of <imageIndex, Image*>

    string dataDir;             // local path to temporary folder
    path filepath;
    int maxIndex;               // maximum index in the imagetable (need it to assign a unique new index)

    struct timeval time_start,time_end;

    //! Given an image use the dictionary to run majority voting...
    int getVotes(Image *img, PointMatchMap &fm);

    //! Processes candidate images and eliminate the ones that do not satify the geometric constraints
    //! @forceSimilarImages: if set to true, will insert the image even if there is a very similar one in the map
    bool enforceGeometricConstraints(PointMatchMap &matches, CandidateList &candidates, bool forceSimilarImages=false);

    // @keepdata: to keep raw image in the memory for quick access
    vertexDescriptor addImageAsVertex(Image *img, bool keepdata=false);

    int chooseIndex(int i);

    void cleanImageList();

    bool writeImageList(string filename);
    bool readImageList(string filename);

    bool saveImageToFile(Image *img, string path);

    vector<Image*> trackingList;

public:

    SimMap();
    SimMap(path filename);
    ~SimMap();

    SimGraph* getGraph() { return sg; }
    Dictionary* getDictionary() { return dict; }
    ImageTable* getImages() { return &imageLookup; }
    string getDataDir() { return dataDir; }

    void clearTrackingList() { trackingList.clear(); }

    bool isEmpty();

    void setFilepath(path s) { filepath = s; }
    path getFilepath() { return filepath; }
    string getFilename() { return filepath.stem(); }

    void test();
    void test(QString f1,QString f2);

    // returns the assigned indices of images inserted (there may be more than 1 image insertions if backtracking is on)
    vector<int> addImage(QString imagefile, int indx = -1, bool force=false, int trackingMode=0, bool addEdge=true); // read image file and create an Image, then same as addImage(Image *)
    vector<int> addImage(Image *img, bool force=false, int trackingMode=0, bool addEdge=true); // adds image to the map if it needs to be added
    //! Localizes the image in the map and returns localized imageindex>0
    int localizeImage(Image *img);

    bool addEdge(int v1, int v2);
    bool removeVertex(int v);
    bool removeEdge(int v1, int v2);
    bool removeEdges(vector<pair<int,int> > lim1, vector<pair<int,int> > lim2);
    bool removeConnectedComp(int c);

    double scoreMap();

    bool writeMap(path filepath, bool writeIndvFiles=false);
    bool readMap(path filepath);

    void printGraph(QString &imagefile, QString imageformat, QString graphlayout);

    /* STATISTICS */
    void updateStats() { sg->updateStats(); }
    void writeStats(QString filepath);
    int numVertices() { return sg->numVertices(); }
    int numEdges() { return sg->numEdges(); }
    int graphMinDegree() { return sg->minDegree; }
    int graphMaxDegree() { return sg->maxDegree; }
    double graphAvgDegree() { return sg->avgDegree; }
    int graphNumConnComp() { return sg->numConnComp; }


    // temporary function to transfer data to Matlab
    void dumpGraphHeader(ofstream &out, SimMap *m1, SimMap *m2);
    void dumpGraph(SimMap *m1, SimMap *m2,map<int,int> &oldnewIndex);
};


#endif // SIMMAP_H

